import copy
import itertools
import pandas as pd
import gzip

from pathlib import Path
from collections.abc import Iterable
from collections import defaultdict
from typing import Self, Iterator
from concurrent.futures import ProcessPoolExecutor
from tqdm import tqdm

from rdkit import Chem, DataStructs, Geometry
from rdkit.Chem import Draw, AllChem, rdFMCS, rdDepictor
from rdkit.ML.Cluster import Butina
from rdkit.SimDivFilters.rdSimDivPickers import MaxMinPicker
from PIL import Image

from rdworks import Conf, Mol
from rdworks.view import render_matrix_grid
from rdworks.xml import list_predefined_xml
from rdworks.fingerprint import distance_function, calculate_fingerprints

class MolLibr:
    def __init__(self, 
                 molecules: Iterable | None = None,
                 names: Iterable | None = None,
                 std:bool=False,
                 max_workers:int=4,
                 chunksize:int=10,
                 progress:bool=False) -> None:
        """Create a rdworks.MolLibr object.

        Args:
            molecules (Iterable | None, optional): a list/tuple/set of molecules 
                (rdworks.Mol | SMILES | rdkit.Chem.Mol). Defaults to None.
            names (Iterable | None, optional): a list/tuple/set of names. Defaults to None.
            std (bool, optional): whether to standardize molecules. Defaults to False.
            max_workers (int, optional): max workers for parallel calculation. Defaults to 4.
            chunksize (int, optional): chunksize for parallel calculation. Defaults to 100.
            progress (bool, optional): whether to show progress bar. Defaults to False.

        Raises:
            ValueError: if counts of molecules and names differ.
            TypeError: if molecule is not rdworks.Mol | SMILES | rdkit.Chem.Mol )
        """
        self.libr = []
        self.max_workers = max_workers
        self.chunksize = chunksize
        self.progress = progress
        self.query = None
        self.threshold = None
        self.clusters = None

        assert isinstance(molecules, Iterable) or molecules is None, "molecules must be iterable or None"
        assert isinstance(names, Iterable) or names is None, "names must be iterable or None"
        
        if isinstance(molecules, Iterable):
            if isinstance(names, Iterable):
                assert len(molecules) == len(names), "molecules and names must be the same counts"
            
            if names is None:
                names = [''] * len(molecules)
            
            for molecular_input, name in zip(molecules, names):
                if isinstance(molecular_input, Mol):
                    _mol = molecular_input
                
                elif isinstance(molecular_input, Chem.Mol) or isinstance(molecular_input, str):
                    _mol = Mol(molecular_input, name=name, std=std)
                
                elif isinstance(molecular_input, Conf):
                    _mol = Mol(molecular_input.rdmol,
                                name=molecular_input.name, 
                                std=std).props.update(molecular_input.props)

                self.libr.append(_mol)
            
            if not any(names):
                self.rename(prefix='entry')
    

    def __str__(self) -> str:
        """Returns string representation.

        Returns:
            str: string representation.
        """
        
        return f"<MolLibr({self.count()})>"
    

    def __iter__(self) -> Iterator:
        """Yields an iterator of molecules.

        Yields:
            Iterator: iterator of molecules.
        """
        return iter(self.libr)
    

    def __next__(self) -> Mol:
        """Next molecule.

        Returns:
            Mol: next molecule.
        """
        return next(self.libr)
    

    def __eq__(self, other: Self) -> bool:
        """Operator `==`.

        Args:
            other (rdworks.MolLibr): other rdworks.MolLibr object.

        Returns:
            Bool: True if other MolLibr object is identical with self.
        """
        if isinstance(other, MolLibr):
            return len(frozenset(self.libr) - frozenset(other.libr)) == 0
        
        return False
        

    def __getitem__(self, index: int | slice) -> Mol | Self:
        """Operator `[]`.

        Args:
            index (Union[int, slice]): index or slice of indexes.

        Returns:
            Mol or MolLibr specified by single index or slice.
        """
        assert self.count() != 0, "library is empty"
        if isinstance(index, slice):
            return MolLibr(self.libr[index])
        else:
            return self.libr[index]
        

    def __setitem__(self, index: int, molecule: Mol) -> Self:
        """Set item.

        Args:
            index (int): index
            molecule (Mol): molecule to replace
        
        Returns:
            Modified self.
        """
        self.libr[index] = molecule

        return self


    def __add__(self, other: Mol | Self) -> Self:
        """Operator `+`.

        Returns a new object, leaving the original objects unchanged (conventional behavior).

        Args:
            other (object): other Mol or MolLibr object.

        Returns:
            A new MolLibr object.
        """
        assert isinstance(other, Mol | MolLibr), "'+' operator expects Mol or MolLibr object"
        
        new_object = self.copy()

        if isinstance(other, Mol):
            new_object.libr.append(other)
        
        elif isinstance(other, MolLibr):
            new_object.libr.extend(other.libr)

        return new_object


    def __iadd__(self, other: Mol | Self) -> Self:
        """Operator `+=`.

        Args:
            other (object): other Mol or MolLibr object.

        Returns:
            modified self.
        """
        assert isinstance(other, Mol | MolLibr), "'+=' operator expects Mol or MolLibr object"

        if isinstance(other, Mol):
            self.libr.append(other)
        
        elif isinstance(other, MolLibr):
            self.libr.extend(other.libr)

        return self


    def __sub__(self, other: Mol | Self) -> Self:
        """Operator `-`.

        Returns a new object, leaving the original objects unchanged (conventional behavior).

        Args:
            other (Mol | MolLibr): other rdworks.Mol or rdworks.MolLibr object.

        Returns:
            A new MolLibr object.
        """
        assert isinstance(other, Mol | MolLibr), "'-' operator expects Mol or MolLibr object"

        if isinstance(other, Mol):
            difference = frozenset(self.libr) - frozenset([other])
        
        elif isinstance(other, MolLibr):
            difference = frozenset(self.libr) - frozenset(other.libr)

        new_object = self.copy()
        new_object.libr = list(difference)

        return new_object
    

    def __isub__(self, other: Mol | Self) -> Self:
        """Operator `-=`.

        Args:
            other (Mol | MolLibr): other molecule or library.

        Returns:
            Modified self.
        """
        assert isinstance(other, Mol | MolLibr), "'-=' operator expects Mol or MolLibr object"

        if isinstance(other, Mol):
            difference = frozenset(self.libr) - frozenset([other])
        
        elif isinstance(other, MolLibr):
            difference = frozenset(self.libr) - frozenset(other.libr)

        self.libr = list(difference)
        
        return self


    def __and__(self, other: Mol | Self) -> Self:
        """Operator `&`.

        Returns a new object, leaving the original objects unchanged (conventional behavior).

        Args:
            other (Mol | MolLibr): other molecule or library.

        Returns:
            A new MolLibr object.
        """
        assert isinstance(other, Mol | MolLibr), "'&' operator expects Mol or MolLibr object"

        if isinstance(other, Mol):
            intersection = frozenset(self.libr) & frozenset([other])

        elif isinstance(other, MolLibr):
            intersection = frozenset(self.libr) & frozenset(other.libr)
        
        new_object = self.copy()
        new_object.libr = list(intersection)
        
        return new_object


    def __iand__(self, other: Mol | Self) -> Self:
        """Operator `&=`.

        Args:
            other (Mol | Self): other molecule or library.

        Returns:
            Modified self.
        """
        assert isinstance(other, Mol | MolLibr), "'&=' operator expects Mol or MolLibr object"

        if isinstance(other, Mol):
            intersection = frozenset(self.libr) & frozenset([other])
        
        elif isinstance(other, MolLibr):
            intersection = frozenset(self.libr) & frozenset(other.libr)

        self.libr = list(intersection)
        
        return self
    

    @staticmethod
    def _mask_similar(mol:Mol, targs:tuple) -> bool:
        """A mask function to return True if molecule is similar with target molecules, `targs`.

        Args:
            mol (Mol): subject rdworks.Mol object.
            targs (tuple): a tuple of rdworks.Mol objects to compare.

        Returns:
            bool: True if molecule is similar with target molecules.
        """
        return mol.is_similar(*targs) # unpack tuple of arguments
    

    @staticmethod
    def _mask_drop(mol:Mol, terms:str | Path) -> bool:
        """A mask function to return True if molecule matches `terms`.

        Note that molecules matching the terms will be dropped (NOT be included) in the compression.

        Args:
            mol (Mol): subject rdworks.Mol object.
            terms (str | Path): rule.

        Returns:
            bool: True if molecule matches the terms.
        """
        return not mol.is_matching(terms)
    
    @staticmethod
    def _map_qed(mol:Mol, properties:list[str]=['QED', 'MolWt', 'LogP', 'TPSA', 'HBD']) -> dict:
        """A map function to apply Mol.qed(`properties`) on `mol`.

        The default behavior of map() is to pass the elements of the iterable to the function by reference. 
        This means that if the function modifies the elements of the iterable, 
        those changes will be reflected in the iterable itself.

        Args:
            mol (Mol): subject rdworks.Mol object.
            properties (list[str], optional): properties. Defaults to ['QED', 'MolWt', 'LogP', 'TPSA', 'HBD'].

        Returns:
            dict: dictionary of properties.
        """
        return mol.qed(properties)


    @staticmethod
    def _mcs_coord_map(subject:Mol, r:Chem.Mol) -> dict:
        s = subject.rdmol
        lcs = rdFMCS.FindMCS([r, s])
        # reference matching indices
        r_indices = r.GetSubstructMatch(lcs.queryMol)
        # subject matching indices
        s_indices = s.GetSubstructMatch(lcs.queryMol)
        # reference matching coordinates (2D)
        r_xy = []
        for i in r_indices:
            pt = r.GetConformer().GetAtomPosition(i)
            r_xy.append(Geometry.Point2D(pt.x, pt.y))
        coord_map = { i : xy for i, xy in zip(s_indices, r_xy) }
        
        return coord_map
    

    @staticmethod
    def _mask_nnp_ready(mol: Mol, model: str) -> bool:
        """A mask function to return True if molecule is NNP ready. 

        Args:
            mol (Mol): rdworks.Mol object.
            model (str): name of NNP model.

        Returns:
            bool: True if molecule is NNP ready.
        """
        return mol.is_nnp_ready(model)
    

    ##################################################
    ### Pipeline Functions (returns Self)
    ##################################################

    def copy(self) -> Self:
        """Returns a copy of self.

        Returns:
            Self: rdworks.MolLibr object.
        """
        return copy.deepcopy(self)
    
    def compute(self, **kwargs) -> Self:
        """Change settings for parallel computing.

        Args:
            max_workers (int, optional): max number of workers. Defaults to 4.
            chunksize (int, optional): chunksize of splitted workload. Defaults to 10.
            progress (bool, optional): whether to show progress bar. Defaults to False.

        Returns:
            Self: rdworks.MolLibr object.
        """
        self.max_workers = kwargs.get('max_workers', self.max_workers)
        self.chunksize = kwargs.get('chunksize', self.chunksize)
        self.progress = kwargs.get('progress', self.progress)
        
        return self


    def rename(self, prefix: str | None = None, sep: str='.', start: int=1) -> Self:
        """Rename molecules with serial numbers in-place and their conformers.
        
        Molecules will be named by a format, `{prefix}{sep}{serial_number}` and 
        conformers will be named accordingly.

        Examples:
            >>> a.rename(prefix='a')

        Args:
            prefix (str, optional): prefix for new name. If prefix is not given and set to None,
                                    molecules will not renamed but conformers will be still renamed. 
                                    This is useful after dropping some conformers and rename them serially.
            sep (str): separator between prefix and serial number (default: `.`)
            start (int): start number of serial number.

        Returns:
            Self: rdworks.MolLibr object.
        """

        num = self.count()
        num_digits = len(str(num)) # ex. '100' -> 3
        if prefix: 
            # use prefix to rename molecules AND conformers
            for (serial, mol) in enumerate(self.libr, start=start):
                if num > 1:
                    serial_str = str(serial)
                    while len(serial_str) < num_digits:
                        serial_str = '0' + serial_str
                    mol.rename(prefix=f"{prefix}{sep}{serial_str}")
                else:
                    mol.rename(prefix)
        else:
            # rename molecules using serial numbers if they have duplicate names
            # name -> name.1, name.2, ...
            count_names = defaultdict(list)
            for idx, mol in enumerate(self.libr):
                count_names[mol.name].append(idx)
            not_unique_names = [name for name, l in count_names.items() if len(l) > 1]
            for idx, mol in enumerate(self.libr):
                if mol.name in not_unique_names:
                    serial = count_names[mol.name].index(idx) + 1
                    mol.rename(f'{mol.name}.{serial}')
            # rename conformers
            for mol in self.libr:
                mol.rename()
        
        return self
    

    def overlap(self, other: Self) -> Self:
        """Returns a common subset with `other` library.

        Args:
            other (Self): rdworks.MolLibr object.

        Returns:
            Self: common subset of rdworks.MolLibr.
        """
        return self.__and__(other)
    

    def similar(self, query: Mol, threshold: float = 0.2, **kwargs) -> Self:
        """Returns a copy of subset that are similar to `query`.

        Args:
            query (Mol): query molecule.
            threshold (float, optional): similarity threshold. Defaults to 0.2.

        Raises:
            TypeError: if query is not rdworks.Mol type.

        Returns:
            Self: a copy of self.
        """
        obj = self.copy().compute(**kwargs)
        
        if isinstance(query, Mol):
            largs = [(query, threshold),] * obj.count()
        else:
            raise TypeError("MolLibr.similar() expects Mol object")
        with ProcessPoolExecutor(max_workers=obj.max_workers) as executor:
            if self.progress:
                mask = list(tqdm(executor.map(MolLibr._mask_similar, obj.libr, largs, chunksize=obj.chunksize),
                    desc="Similar",
                    total=obj.count()))
            else:
                mask = list(executor.map(MolLibr._mask_similar, obj.libr, largs, chunksize=obj.chunksize))
            obj.libr = list(itertools.compress(obj.libr, mask))
        
        return obj



    def unique(self, report=False) -> Self:
        """Removes duplicates and returns a copy of unique library.

        Args:
            report (bool, optional): whether to report duplicates. Defaults to False.

        Returns:
            Self: a copy of self.
        """
        obj = self.copy()
        
        U = {} # unique SMILES
        mask = []
        for mol in obj.libr:
            if mol.smiles in U:
                mask.append(False)
                # ignore the same name or recorded aka
                if (mol.name != U[mol.smiles].name) and (mol.name not in U[mol.smiles].props['aka']):
                    U[mol.smiles].props['aka'].append(mol.name)
            else:
                mask.append(True)
                U[mol.smiles] = mol
        obj.libr = list(itertools.compress(obj.libr, mask))
        if report:
            print("duplicates:")
            for mol in obj.libr:
                if len(mol.props['aka']) > 0:
                    print(f"  {mol.name}({len(mol.props['aka'])}) - {','.join(mol.props['aka'])}")
            print(f"de-duplicated to {obj.count()} molecules")
        
        return obj
    

    def qed(self, properties:list[str]=['QED', 'MolWt', 'LogP', 'TPSA', 'HBD'], **kwargs) -> Self:
        """Returns a copy of self with calculated quantitative estimate of drug-likeness (QED).

        Args:
            properties (list[str], optional): _description_. Defaults to ['QED', 'MolWt', 'LogP', 'TPSA', 'HBD'].

        Returns:
            Self: self.
        """
        self = self.compute(**kwargs)
        lprops = [ properties, ] * self.count()
        with ProcessPoolExecutor(max_workers=self.max_workers) as executor:
            if self.progress:
                self.libr = list(tqdm(
                    executor.map(MolLibr._map_qed, self.libr, lprops, chunksize=self.chunksize),
                    desc="QED Properties",
                    total=self.count()
                    ))
            else:
                self.libr = list(
                    executor.map(MolLibr._map_qed, self.libr, lprops, chunksize=self.chunksize)
                    )
        
        return self
    

    def drop(self, terms: str | Path | None = None, invert: bool = False, **kwargs) -> Self:
        """Drops matched molecules and returns a copy of library with remaining molecules.

        Args:
            terms (str | Path | None, optional): matching terms. Defaults to None.
            invert (bool, optional): whether to invert selection by the `terms`. Defaults to False.

        Returns:
            Self: a copy of self.
        """
        if not terms:
            print(list_predefined_xml())
            return self
        
        obj = self.copy().compute(**kwargs)
        
        lterms = [ terms ] * obj.count()
        with ProcessPoolExecutor(max_workers=obj.max_workers) as executor:
            if obj.progress:
                mask = list(tqdm(
                    executor.map(MolLibr._mask_drop, obj.libr, lterms, chunksize=obj.chunksize),
                    desc="Drop",
                    total=obj.count()))
            else:
                mask = list(
                    executor.map(MolLibr._mask_drop, obj.libr, lterms, chunksize=obj.chunksize))
            if invert:
                mask = [not b for b in mask]
            obj.libr = list(itertools.compress(obj.libr, mask))
        
        return obj
    

    def pick_diverse(self, n:int, seed:int=0, fp_type:str='morgan', radius:int=2, n_bits:int=2048) -> Self:
        """
        Select diverse molecules using MaxMin algorithm.
        
        Args:
            n (int): Number of molecules to select.
            seed (int, optional): Random seed for reproducibility. Defaults to 0.
            fp_type (str, optional): Type of fingerprint to use: morgan, rdkit, maccs. Defaults to 'morgan'
            radius (int, optional): Radius for Morgan fingerprints. Defaults to 2
            n_bits (int, optional): Number of bits for fingerprints. Defaults to 2048
        
        Returns:
            List of selected indices
        """
        # Calculate fingerprints
        fps = calculate_fingerprints(self.libr, fp_type, radius, n_bits)
        
        # Remove None values (invalid molecules)
        valid_indices = [i for i, fp in enumerate(fps) if fp is not None]
        valid_fps = [fp for fp in fps if fp is not None]
        
        if len(valid_fps) < n:
            raise ValueError(f"Requested {n} molecules but only {len(valid_fps)} valid")
        
        # Create distance function
        def dist_func(i, j):
            return distance_function(i, j, valid_fps)
        
        # Run MaxMin picker
        picker = MaxMinPicker()
        picks = picker.LazyBitVectorPick(dist_func, len(valid_fps), n, seed=seed)
        
        original_indices = [valid_indices[i] for i in picks]
        mask = [True for i in range(self.count()) if i in original_indices]
        
        self.libr = list(itertools.compress(self.libr, mask))

        return self


    def align_drawing(self, 
                      ref: int = 0, 
                      mcs: bool = True, 
                      scaffold: str = "", 
                      coordgen:bool = True, 
                      **kwargs,
                      ) -> Self:
        """Align 2D drawings by using MCS or scaffold SMILES.

        Args:
            ref (int, optional): index to the reference. Defaults to 0.
            mcs (bool, optional): whether to use MCS(maximum common substructure). Defaults to True.
            scaffold (str, optional): whether to use scaffold (SMILES). Defaults to "".

        Returns:
            Self: self
        """

        obj = self.copy().compute(**kwargs)
        
        if scaffold:
            # scaffold (SMILES) of the reference 2D drawing
            ref_2d_rdmol = Chem.MolFromSmiles(scaffold)
        else:
            # maximum common substructure to the reference 2D drawing
            assert ref >=0 and ref < obj.count(), f"ref should be [0,{obj.count()-1}]"
            ref_2d_rdmol = obj.libr[ref].rdmol

        rdDepictor.SetPreferCoordGen(coordgen)
        rdDepictor.Compute2DCoords(ref_2d_rdmol)
        # AllChem.Compute2DCoords(ref_2d_rdmol)
        
        with ProcessPoolExecutor(max_workers=obj.max_workers) as executor:
            if obj.progress:
                coord_maps = list(tqdm(
                    executor.map(MolLibr._mcs_coord_map, 
                        obj.libr, # subject
                        itertools.repeat(ref_2d_rdmol), # infinite iterator
                        chunksize=obj.chunksize),
                    desc="align drawingp",
                    total=obj.count()))
            else:
                coord_maps = list(
                    executor.map(MolLibr._mcs_coord_map,
                        obj.libr, # subject
                        itertools.repeat(ref_2d_rdmol), # infinite iterator
                        chunksize=obj.chunksize))
        
        for mol, coord_map in zip(obj.libr, coord_maps):
            rdDepictor.Compute2DCoords(mol.rdmol, coordMap=coord_map)
            # AllChem.Compute2DCoords(mol.rdmol, coordMap=coord_map)


        # for idx, mol in enumerate(obj.libr):
        #     if mcs and idx == ref:
        #         continue

        #     # largest common substructure
        #     lcs = rdFMCS.FindMCS([ref_2d_rdmol, mol.rdmol])

        #     # matching indices
        #     ref_xy_coords = []
        #     for i in ref_2d_rdmol.GetSubstructMatch(lcs.queryMol):
        #         pt = ref_2d_rdmol.GetConformer().GetAtomPosition(i)
        #         ref_xy_coords.append(Geometry.Point2D(pt.x, pt.y))
        #     sub_indices = mol.rdmol.GetSubstructMatch(lcs.queryMol)
        #     coord_map = { i : xy for i, xy in zip(sub_indices, ref_xy_coords) }
        #     AllChem.Compute2DCoords(mol.rdmol, coordMap=coord_map)

        return obj


    def nnp_ready(self, model: str, **kwargs) -> Self:
        """Returns a copy of subset of library that is ready to given neural network potential.

        Examples:
            >>> libr = rdworks.MolLibr(drug_smiles, drug_names)
            >>> ani2x_compatible_subset = libr.nnp_ready('ANI-2x', progress=False)

        Args:
            model (str): name of model.

        Returns:
            Self: subset of library.
        """
        obj = self.copy().compute(**kwargs)
        lmodel = [model,] * self.count()
        with ProcessPoolExecutor(max_workers=obj.max_workers) as executor:
            if obj.progress:
                mask = list(tqdm(
                    executor.map(self._mask_nnp_ready, obj.libr, lmodel, chunksize=obj.chunksize),
                    desc="NNP ready",
                    total=obj.count()))
            else:
                mask = list(
                    executor.map(self._mask_nnp_ready, obj.libr, lmodel, chunksize=obj.chunksize))
            obj.libr = list(itertools.compress(obj.libr, mask))
        
        return obj


    ##################################################
    ### endpoints
    ##################################################

    def count(self) -> int:
        """Returns number of molecules.

        Returns:
            int: count of molecules.
        """
        return len(self.libr)
    

    def cluster(self, 
                threshold: float = 0.3, 
                ordered: bool = True, 
                drop_singleton: bool = True,
                ) -> list:
        """Clusters molecules using fingerprint.

        Args:
            threshold (float, optional): Tanimoto similarity threshold. Defaults to 0.3.
            ordered (bool, optional): order clusters by size of cluster. Defaults to True.
            drop_singleton (bool, optional): exclude singletons. Defaults to True.

        Returns:
            list: [(centroid_1, idx, idx,), (centroid_2, idx, idx,), ...]
        """
        for mol in self.libr:
            if not mol.fp:
                mol.fp = mol.MFP2.GetFingerprint(mol.rdmol)
        fps = [ mol.fp for mol in self.libr if mol.fp ]
        n = len(fps)
        # first generate the distance matrix:
        dmat = []
        for i in range(1, n):
            sims = DataStructs.BulkTanimotoSimilarity(fps[i], fps[:i])
            dmat.extend([1-x for x in sims])
        # Butina hierarchical clustering:
        # clusters is a list of list of indices
        clusters = Butina.ClusterData(dmat, 
                                      nPts=n, 
                                      distThresh=threshold, 
                                      isDistData=True, 
                                      reordering=True)
        if ordered:
            # in the order of cluster size, from the largest to the smallest
            clusters = sorted(clusters, key=lambda indices: len(indices), reverse=True)
        
        if drop_singleton:
            clusters = [indices for indices in clusters if len(indices) > 1]
        
        return clusters
    
    
    def to_sdf(self, 
                path: str | Path, 
                confs: bool = False, 
                props: bool = True, 
                separate: bool = False,
                ) -> None:
        """Writes to .sdf or .sdf.gz file.

        Chem.SDWriter is supposed to write all non-private molecular properties.

        `dirname/filename.sdf` -> `dirname/filename_{molecule name}.sdf`
        `dirname/filename.sdf.gz` -> `dirname/filename_{molecule name}.sdf.gz`

        Args:
            path (str or PosixPath) : output filename or path
            confs (bool) : whether to write 3D coordinates and conformer properties. Defaults to False.
            props (bool) : whether to write SDF properties. Defaults to True.
            separate (bool) : write each molecule to separate files. Defaults to False.
        """
        if isinstance(path, str):
            path = Path(path)
        # PurePosixPath('my/dir/mol.sdf.gz').suffix -> '.gz'
        # PurePosixPath('my/dir/mol.sdf.gz').suffixes -> ['.sdf', '.gz']
        # PurePosixPath('my/dir/mol.sdf').name -> 'mol.sdf'
        # PurePosixPath('my/dir/mol.sdf').with_name('mol2.sdf') -> PurePath('my/dir/mol2.sdf')
        suffix = path.suffix 
        suffixes = ''.join(path.suffixes) 
        prefix = path.name.replace(suffixes, '')
        if separate:
            for mol in self.libr:
                if suffix == '.gz':
                    with gzip.open(path.with_name(f'{prefix}_{mol.name}.sdf.gz'), "wt") as f:
                        f.write(mol.to_sdf(confs, props))
                else:
                    with open(path.with_name(f'{prefix}_{mol.name}.sdf'), "w") as f:
                        f.write(mol.to_sdf(confs, props))

        else:
            if suffix == '.gz':
                with gzip.open(path, "wt") as f:
                    for mol in self.libr:
                        f.write(mol.to_sdf(confs, props))
            else:
                with open(path, "w") as f:
                    for mol in self.libr:
                        f.write(mol.to_sdf(confs, props))


    def to_smi(self, path: str | Path) -> None:
        """Writes to .smi file.

        Args:
            path (str | Path): output filename or path.
        """
        if isinstance(path, Path):
            path = path.as_posix() # convert to string
        if path.endswith('.gz'):
            with gzip.open(path, "wt") as smigz:
                for mol in self.libr:
                    smigz.write(f'{mol.smiles} {mol.name}\n')
        else:
            with open(path, "w") as smi:
                for mol in self.libr:
                    smi.write(f'{mol.smiles} {mol.name}\n')


    def to_svg(self, 
               mols_per_row: int = 5,
               width: int = 200, 
               height: int = 200,
               atom_index: bool = False,
               redraw: bool = False,
               coordgen: bool = False) -> str:
        """Writes to a .svg strings for Jupyter notebook.

        Args:
            path (str | Path): output filename or path.
            mols_per_row (int, optional): number of molecules per row. Defaults to 5.
            width (int, optional): width. Defaults to 200.
            height (int, optional): height. Defaults to 200.
            atom_index (bool, optional): whether to show atom index. Defaults to False.
            redraw (bool, optional): whether to redraw. Defaults to False.
            coordgen (bool, optional): whether to use coordgen. Defaults to False.
        """
        
        rdmols  = [mol.rdmol for mol in self.libr]
        legends = [mol.name for mol in self.libr]

        svg_string = render_matrix_grid(rdmols, 
                                        legends,
                                        mols_per_row = mols_per_row, 
                                        width = width, 
                                        height = height,
                                        atom_index = atom_index,
                                        redraw = redraw,
                                        coordgen = coordgen,
                                        svg = True,
                                        )
        
        return svg_string


    def to_png(self,
               filename: str | Path | None = None,
               mols_per_row: int = 5,
               width: int = 200, 
               height: int = 200,
               atom_index: bool = False,
               redraw: bool = False,
               coordgen: bool = False,
               ) -> Image.Image | None:
        """Writes to a .png file.

        Args:
            mols_per_row (int, optional): number of molecules per row. Defaults to 5.
            width (int, optional): width. Defaults to 200.
            height (int, optional): height. Defaults to 200.
            atom_index (bool, optional): whether to show atom index. Defaults to False.
            redraw (bool, optional): whether to redraw. Defaults to False.
            coordgen (bool, optional): whether to use coordgen. Defaults to False.
        """
        rdmols  = [mol.rdmol for mol in self.libr]
        legends = [mol.name for mol in self.libr]

        img = render_matrix_grid(rdmols, 
                                legends,
                                mols_per_row = mols_per_row, 
                                width = width, 
                                height = height,
                                atom_index = atom_index,
                                redraw = redraw,
                                coordgen = coordgen,
                                svg = False,
                                )
        
        if filename is None:
            return img
        else:
            if isinstance(filename, Path):
                filename = filename.as_posix()
            img.save(filename)


    def to_html(self) -> str:
        """Writes to HTML strings.

        Returns:
            str: HTML strings.
        """
        HTML = "<html><body>"
        for mol in self.libr:
            HTML += mol.to_html(htmlbody=False)
        HTML += "</body></html>"
        return HTML


    def to_df(self, 
            name: str = 'name', 
            smiles: str = 'smiles', 
            confs: bool = False,
            ) -> pd.DataFrame:
        """Returns a Pandas DataFrame.

        Args:
            name (str, optional): column name for name. Defaults to 'name'.
            smiles (str, optional): column name for SMILES. Defaults to 'smiles'.
            confs (bool, optional): whether to include conformer properties. Defaults to False.

        Returns:
            pd.DataFrame: pandas DataFrame.
        """
        if confs:
            exclude = ['coord']
            property_columns = set()
            for mol in self.libr:
                for conf in mol.confs:
                    for k in conf.props:
                        if k not in exclude:
                            property_columns.add(k)
            property_columns = property_columns - set([name, smiles])
            data = {name:[], smiles:[]}
            data.update({k:[] for k in property_columns})
            for mol in self.libr:
                for conf in mol.confs:
                    data[name].append(conf.name)
                    data[smiles].append(mol.smiles)
                    for k in property_columns:
                        if k in conf.props:
                            data[k].append(conf.props[k])
                        else:
                            data[k].append(None)
        else:
            property_columns = set()
            for mol in self.libr:
                for k in mol.props:
                    property_columns.add(k)
            property_columns = property_columns - set([name, smiles])
            data = {name:[], smiles:[]}
            data.update({k:[] for k in property_columns})
            for mol in self.libr:
                data[name].append(mol.name)
                data[smiles].append(mol.smiles)
                for k in property_columns:
                    if k in mol.props:
                        data[k].append(mol.props[k])
                    else:
                        data[k].append(None)
        
        return pd.DataFrame(data)


    def to_csv(self, 
               path: str | Path, 
               confs: bool = False, 
               decimals:int = 3,
               ) -> None:
        """Writes to a .csv file.

        Args:
            path (str | Path): output filename or path.
            confs (bool, optional): whether to include conformer properties. Defaults to False.
            decimals (int, optional): decimal places for float numbers. Defaults to 3.
        """
        df = self.to_df(confs=confs)
        df.to_csv(path, index=False, float_format=f'%.{decimals}f')


    def to_batches(self, batchsize_atoms: int = 1000) -> list:
        """Split workload flexibily into a numer of batches.

        - Each batch has up to `batchsize_atoms` number of atoms.
        - Conformers originated from a same molecule can be splitted into multiple batches.
        - Or one batch can contain conformers originated from multiple molecules.

        coord: coordinates of input molecules (N, m, 3) where N is the number of structures and
        m is the number of atoms in each structure.
        numbers: atomic numbers in the molecule (include H). (N, m)
        charges: (N,)

        Args:
            batchsize_atoms: max. number of atoms in a batch.

        Returns:
            list: list of batches.
        """

        pre_batches = []
        batch_confs = []
        batch_mols = []
        batch_n_atoms = 0

        for mol in self.libr:
            for conf in mol.confs:
                n_atoms = conf.props['atoms']
                if (batch_n_atoms + n_atoms) > batchsize_atoms:
                    pre_batches.append((batch_mols, batch_confs, batch_n_atoms))
                    # start over a new batch
                    batch_mols =  [mol]
                    batch_confs = [conf]
                    batch_n_atoms = n_atoms
                else:
                    batch_mols.append(mol)
                    batch_confs.append(conf)
                    batch_n_atoms += n_atoms
        
        if batch_n_atoms > 0: # last remaining batch
            pre_batches.append((batch_mols, batch_confs, batch_n_atoms))
        
        batches = []
        
        for i, (batch_mols, batch_confs, batch_n_atoms) in enumerate(pre_batches, start=1):
            charges = [mol.props['charge'] for mol in batch_mols]
            coord = [conf.rdmol.GetConformer().GetPositions().tolist() for conf in batch_confs]
            # to be consistent with legacy code
            coord = [[tuple(xyz) for xyz in inner] for inner in coord]
            # numbers should be got from conformers because of hydrogens
            numbers = [[a.GetAtomicNum() for a in conf.rdmol.GetAtoms()] for conf in batch_confs]
            batches.append((coord, numbers, charges, batch_confs, batch_mols))
        
        return batches


"""
@author: Roy Gonzalez-Aleman                            [roy_gonzalez@fq.uh.cu]
@author: Daniel Platero Rochart                      [daniel.platero@gmail.com]
"""
from collections import deque, OrderedDict
from typing import Tuple, Optional

import numpy as np
import pandas as pd

from bitarray import util as bu
from bitarray import bitarray as ba

from ..autograph.centroid import centroid_medoid



def convert_to_bitarr_matrix(rmsdMatrix:np.array, cutoff:float) -> OrderedDict:
    """
    Convert RMSD binary-encoded square matrix.
    Pairwise similarity is saved in RAM as bits (dict of bitarrays), not floats.
    Returns:
        matrix : collections.OrderedDict. dict of bitarrays.
    """
    N = rmsdMatrix.shape[0]
    cutoff = np.full(N, cutoff, dtype=np.float32)
    # numpy.full(shape, fill_value, dtype=None, order='C', *, like=None)
    # Return a new array of given shape and type, filled with fill_value.
    matrix = OrderedDict()
    to_explore = range(N)
    for i in to_explore:
        rmsd_ = rmsdMatrix[i,:]
        # mdtraj.rmsd(target, reference, frame=0, atom_indices=None, parallel=True, precentered=False)
        # Compute RMSD of all conformations in target to a reference conformation. Note, this will center the conformations in place.
        vector_np = np.less_equal(rmsd_, cutoff)
        # Return the truth value of (x1 <= x2) element-wise.
        bitarr = ba()
        bitarr.pack(vector_np.tobytes())
        bitarr.fill()
        matrix.update({i: bitarr})
    return matrix



def calc_matrix_degrees(unclustered_bit, matrix):
    """
    Calculate number of neighbors (degree) of unclustered nodes in matrix.

    Parameters
    ----------
    unclustered_bit : bitarray.bitarray
        bitarray with indices of unclustered nodes turned on.
    matrix : collections.OrderedDict
        dict of bitarrays.

    Returns
    -------
    degrees : numpy.ndarray
        array containing each node degree. Clustered nodes have degree = 0.

    """
    one = ba('1')
    degrees = np.zeros(len(unclustered_bit), dtype=np.int32)
    for node in unclustered_bit.itersearch(one):
        try:
            degrees[node] = matrix[node].count()
        except KeyError:
            pass
    return degrees


def colour_matrix(degrees, matrix):
    """
    Greedy coloring of bit-encoded RMSD matrix.

    Parameters
    ----------
    degrees : numpy.ndarray
        array containing each node degree. Clustered nodes have degree = 0.
    matrix : collections.OrderedDict
        dict of bitarrays.

    Returns
    -------
    colors : numpy.ndarray
        array of colors assigned to each node of the matrix.
    """
    # Constants ---------------------------------------------------------------
    N = degrees.size
    m = len(matrix)
    one = ba('1')
    xcolor = 0
    # Initialize containers ---------------------------------------------------
    ordered_by_degrees = iter((-degrees[:m]).argsort())
    colors = np.zeros(N, dtype=np.int32)
    colored = ba(N)
    colored.setall(0)
    seen = set()
    while True:
        # Retrieve the max-degree node ----------------------------------------
        max_node = next(ordered_by_degrees)
        if max_node in seen:
            continue
        seen.add(max_node)
        xcolor += 1
        not_neighbors = ~ matrix[max_node]
        not_colored = ~colored
        candidates = not_neighbors & not_colored
        # Nodes passing conditions (not-neighb, not-colored, not-neighb) ------
        passed = [max_node]
        for candidate in candidates.itersearch(one):
            passed.append(candidate)
            try:
                candidates &= ~matrix[candidate]
            except KeyError:
                continue
            if not candidates.any():
                break
        seen.update(passed)
        # Deliver a color class to passed nodes -------------------------------
        colors[passed] = xcolor
        colored = ba()
        colored.pack(colors.astype(np.bool_).tobytes())
        if colored.count(0) == 0:
            break
    return colors


def bitarray_to_np(bitarr):
    """
    Convert from bitarray.bitarray to numpy.ndarray efficiently.

    Parameters
    ----------
    bitarr : bitarray.bitarray
        a bitarray.

    Returns
    -------
    numpy.ndarray
        boolean bitarray equivalent to the binary bitarray input object.
    """
    return np.unpackbits(bitarr).astype(np.bool_)


def do_bit_cascade(big_node, degrees, colors, matrix, max_):
    """
    Perform succesive AND operations between an initial bitarray and subsequent
    bitarray candidates to search for a clique.

    Parameters
    ----------
    big_node : int
        node whose bitarray will start the operations.
    degrees : numpy.ndarray
        array containing each node degree. Clustered nodes have degree = 0.
    colors : numpy.ndarray
        array of colors assigned to each node of the matrix.
    clustered_bit : bitarray.bitarray
        bitarray with indices of clustered nodes turned on.
    matrix : collections.OrderedDict
        dict of bitarrays.
    max_ : int
        Stop iterative AND operations after the initial bitarray has max_
        bits turned on.

    Returns
    -------
    init_cascade : bitarray.bitarray
        initial bitarray before any AND operation.
    ar : numpy.ndarray
        array of nodes forming a clique.
    """
    init_cascade = matrix[big_node]
    # .... recovering neighbors and their information .........................
    neighb = bitarray_to_np(init_cascade).nonzero()[0]
    neighb_colors = colors[neighb]
    if len(set(neighb_colors.tolist())) <= max_:
        return None
    neighb_degrees = degrees[neighb]
    g = np.bincount(neighb_colors)
    neighb_g = g[neighb_colors]
    # .... ordering neighbors by g ---> colors ---> degrees ...................
    idx = np.lexsort([-neighb_degrees, neighb_colors, neighb_g])
    candidates_info = zip(neighb[idx], neighb_colors[idx])

    # .... BitCascade considering divergence ..................................
    counter = 0
    seen = set()
    for candidate, color in candidates_info:
        if (color in seen) or (not init_cascade[candidate]):
            continue
        seen.add(color)
        init_cascade = matrix[candidate] & init_cascade
        counter += 1
        COUNT = init_cascade.count()
        if (COUNT <= max_):
            return None
        if counter >= COUNT:
            break
    ar = np.nonzero(np.unpackbits(init_cascade).astype(np.bool_))[0]
    return init_cascade, ar


def set_to_bitarray(set_, N):
    """
    Convert from python set to bitarray.bitarray.

    Parameters
    ----------
    set_ : set
        a python set.
    N : int
        lenght of the desired bitarray. It must be greater than the maximum
        value of indices present in set.

    Returns
    -------
    bitarr : bitarray.bitarray
        bitarray of lenght N with indices present in set turned on.
    """
    zero_arr = np.zeros(N, dtype=np.bool_)
    zero_arr[list(set_)] = 1
    bitarr = ba()
    bitarr.pack(zero_arr.tobytes())
    return bitarr


def get_cluster_stats(clusters):
    """
    Get "cluster_statistics.txt" containing clusterID, cluster_size, and
    cluster percentage from trajectory.

    Parameters
    ----------
    clusters : numpy.ndarray
        array of clusters ID.
    outdir : str
        Path where to create the VMD visualization .log.

    Returns
    -------
    clusters_df : pandas.DataFrame
        dataframe with cluster_statistics info.
    """
    clusters_df = pd.DataFrame(columns=['cluster_id', 'size', 'percent'])
    clusters_df['cluster_id'] = list(range(0, clusters.max() + 1))
    sizes = []
    for x in clusters_df.cluster_id:
        sizes.append(len(np.where(clusters == x)[0]))
    clusters_df['size'] = sizes

    sum_ = clusters_df['size'].sum()
    percents = [round(x / sum_ * 100, 4) for x in clusters_df['size']]
    clusters_df['percent'] = percents

    return clusters_df


def BitQT(rmsdMatrix:np.array, cutoff:float, min_clust_size:int=2, nclust:Optional[int]=None) -> Tuple:
    """BitQT clustering
    Returns:
        (cluster_assignment, centroid_indices)
    """
    matrix = convert_to_bitarr_matrix(rmsdMatrix, cutoff)
    # ++++ Tracking clust/uNCLUSTERed bits to avoid re-computations +++++++++++
    N = len(matrix[0])
    m = len(matrix)
    unclust_bit = ba(N)
    unclust_bit.setall(1)
    clustered_bit = unclust_bit.copy()
    clustered_bit.setall(0)
    zeros = np.zeros(N, dtype=np.int32)
    # ++++ Save clusters in an array (1 .. N) +++++++++++++++++++++++++++++++++
    clusters_array = np.zeros(N, dtype=np.int32)
    NCLUSTER = 0
    clustered = set()
    nmembers = []
    # ++++ Coloring ordered vertices (1 .. N) +++++++++++++++++++++++++++++++++
    degrees = calc_matrix_degrees(unclust_bit, matrix)
    ordered_by_degs = degrees.argsort()[::-1]
    colors = colour_matrix(ordered_by_degs, matrix)
    # colors[np.frombuffer(clustered_bit.unpack(), dtype=np.bool)] = 0

    # =========================================================================
    # 2. Main algorithm: BitQT !
    # =========================================================================
    while any(degrees):
        NCLUSTER += 1
        # ++++ Find a big clique early ++++++++++++++++++++++++++++++++++++++++
        big_node = degrees.argmax()
        bit_clique, big_clique = do_bit_cascade(big_node, degrees, colors,
                                                matrix, 0)
        big_clique_size = big_clique.size
        # ++++ Find promising nodes +++++++++++++++++++++++++++++++++++++++++++
        biggers = degrees > big_clique_size
        biggers[big_clique] = False
        cluster_colors = colors[big_clique]
        biggers_colors = colors[biggers]
        promising_colors = np.setdiff1d(biggers_colors, cluster_colors)
        promising_nodes = deque()
        for x in promising_colors:
            promising_nodes.extend(((colors == x) & biggers).nonzero()[0])
        # ++++ Explore all promising nodes ++++++++++++++++++++++++++++++++++++
        cum_found = big_clique
        while promising_nodes:
            node = promising_nodes.popleft()
            try:
                bit_clique, clique = do_bit_cascade(node, degrees, colors,
                                                    matrix, big_clique_size)
                CLIQUE_SIZE = len(clique)
            except TypeError:
                CLIQUE_SIZE = 0
            # ++++ Cumulative update only if biggers candidates are found +++++
            if CLIQUE_SIZE > big_clique_size:
                big_node = node
                big_clique = clique
                big_clique_size = big_clique.size
                # ++++ Repeat previous condition ++++++++++++++++++++++++++++++
                cum_found = np.concatenate((cum_found, big_clique))
                biggers = degrees > big_clique_size
                biggers[cum_found] = False
                cluster_colors = colors[big_clique]
                biggers_colors = colors[biggers]
                promising_colors = np.setdiff1d(biggers_colors, cluster_colors)
                promising_nodes = deque()
                for x in promising_colors:
                    promising_nodes.extend(((colors == x) & biggers).nonzero()[0])
        nmembers.append(big_clique_size)

        if (big_clique_size < min_clust_size) or (nclust and NCLUSTER == nclust):
            break

        # ++++ Save new cluster & update NCLUSTER +++++++++++++++++++++++++++++
        clusters_array[big_clique] = NCLUSTER
        # ++++ Update (un)clustered_bit +++++++++++++++++++++++++++++++++++++++
        clustered.update(big_clique)
        clustered_bit = set_to_bitarray(clustered, N)
        unclust_bit = ~clustered_bit
        # ++++ Hard erasing of clustered frames from matrix +++++++++++++++++++
        degrees = zeros.copy()
        for x in unclust_bit[:m].itersearch(ba('1')):
            degrees[x] = matrix[x].count()
            if bu.count_and(matrix[x], clustered_bit):
                matrix[x] &= (matrix[x] ^ clustered_bit)

    # =========================================================================
    # 3. Output
    # =========================================================================
    # cluster_stats = get_cluster_stats(clusters_array[:m], args.outdir)

    cluster_assignment = list(clusters_array[:m])
    centroid_indices = centroid_medoid(cluster_assignment, rmsdMatrix)

    return cluster_assignment, centroid_indices


import copy
import itertools
import json
import logging
import tempfile
import os

from io import StringIO, BytesIO
from pathlib import Path
from collections import defaultdict, namedtuple
from collections.abc import Callable
from typing import Iterator, Self

import numpy as np
# import pandas as pd
# import matplotlib.ticker as ticker
# import matplotlib.pyplot as plt
# import seaborn as sns

from spyrmsd.molecule import Molecule as spyrmsd_Molecule
from spyrmsd.rmsd import rmsdwrapper as spyrmsd_wrapper
from spyrmsd.rmsd import symmrmsd as spyrmsd_symmrmsd

import CDPL
import CDPL.Chem
import CDPL.ConfGen

from rdkit import Chem, DataStructs
from rdkit.Chem import ( 
    rdMolDescriptors, AllChem, Descriptors, QED, 
    rdFingerprintGenerator,
    Draw, rdDepictor, inchi,
    rdDistGeom, rdMolAlign, rdMolTransforms, rdmolops
    )
from rdkit.ML.Cluster import Butina


from PIL import Image

from rdworks.conf import Conf, batch_optimize, batch_singlepoint
from rdworks.std import generate_inchi_key, desalt_smiles, standardize, clean_2d
from rdworks.torsion import get_torsion_angle_atom_indices
from rdworks.xml import list_predefined_xml, get_predefined_xml, parse_xml
from rdworks.scaffold import rigid_fragment_indices
from rdworks.descriptor import rd_descriptor, rd_descriptor_f
from rdworks.utils import convert_tril_to_symm, QT, recursive_round, compress_string, decompress_string
from rdworks.units import ev2kcalpermol
from rdworks.autograph import NMRCLUST, DynamicTreeCut, RCKmeans, AutoGraph
from rdworks.bitqt import BitQT
from rdworks.view import render_svg, render_png
from rdworks.stereoisomers import enumerate_stereoisomers, enumerate_ring_bond_stereoisomers
from rdworks.torsion import get_torsion_angle_atom_indices
import rdworks.utils

from scour.scour import scourString


logger = logging.getLogger(__name__)


class Mol:
    """Container for molecular structure, conformers, and other information."""

    MFP2 = rdFingerprintGenerator.GetMorganGenerator(radius=2, fpSize=2048)
    
    ETKDG_params = rdDistGeom.ETKDGv3()
    ETKDG_params.useSmallRingTorsions = True
    ETKDG_params.maxIterations = 2000
        
    def __init__(self, 
                 molecule: str | Chem.Mol | Conf | None = None, 
                 name: str = '', 
                 std: bool = False,
                 reset_isotope: bool = True,
                 remove_H: bool = True,
                 max_workers: int = 1,
                 chunksize: int = 4,
                 progress: bool = False) -> None:
        """Initialize.

        Examples:
            >>> from rdworks import Mol
            >>> m = Mol('c1ccccc1', name='benzene')
            
        Args:
            molecule (str | Chem.Mol | None): SMILES or rdkit.Chem.Mol or None
            name (str): name of the molecule. Defaults to ''.
            std (bool): whether to standardize the molecule. Defaults to False.
            max_workers (int): number of maximum workers for parallelization. Defaults to 1.
            chunksize (int): batch size for parallelization. Defaults to 4.
            progress (bool): whether to show progress bar. Defaults to False.
        """
        assert isinstance(molecule, str | Chem.Mol | Conf) or molecule is None

        self.name = ''
        self.rdmol = None # 2D, one and only one Conformer
        self.smiles = '' # isomeric SMILES
        self.confs = [] # container for 3D conformers
        self.InChIKey = '' # 27 characters (SHA-256 hash of InChI)
        self.props = {}
        
        self.max_workers = max_workers
        self.chunksize = chunksize
        self.progress = progress
        self.is_confs_aligned = False  
        self.fp = None

        if molecule is None:
            return

        if isinstance(molecule, str): # 1-D SMILES           
            try:
                if "." in molecule: # mandatory desalting
                    (self.smiles, self.rdmol) = desalt_smiles(molecule)
                else:
                    self.rdmol = Chem.MolFromSmiles(molecule)
                    self.smiles = Chem.MolToSmiles(self.rdmol)
            except:
                raise ValueError(f'Mol() Error: invalid SMILES {molecule}')
 
        elif isinstance(molecule, Chem.Mol): # 2-D or 3-D Chem.Mol
            try:
                self.rdmol, _ = clean_2d(molecule, reset_isotope, remove_H)
                self.smiles = Chem.MolToSmiles(self.rdmol)
                self.confs = [Conf(x) for x in _]
            except:
                raise ValueError(f'Mol() Error: invalid Chem.Mol object')
            
        elif isinstance(molecule, Conf): # 3-D input
            try:
                self.rdmol, _ = clean_2d(molecule.rdmol, reset_isotope, remove_H)
                self.smiles = Chem.MolToSmiles(self.rdmol)
                self.confs = [molecule]
            except:
                raise ValueError(f'Mol() Error: invalid Conf object')

        try:
            if std:
                self.rdmol = standardize(self.rdmol)
                self.smiles = Chem.MolToSmiles(self.rdmol)
        except:
            raise RuntimeError('Mol() Error: standardization')
    
        assert self.smiles and self.rdmol, "Mol() Error: invalid molecule"

        rdDepictor.Compute2DCoords(self.rdmol)

        try:
            self.name = str(name)
        except:
            self.name = 'untitled'

        self.rdmol.SetProp('_Name', self.name) # _Name can't be None
        self.InChIKey = generate_inchi_key(self.rdmol)
        self.props.update({
            'aka' : [], # <-- to be set by MolLibr.unique()
            'atoms' : self.rdmol.GetNumAtoms(), # hydrogens not excluded?
            'charge': rdmolops.GetFormalCharge(self.rdmol),
            "nrb" : Descriptors.NumRotatableBonds(self.rdmol),
            })


    def __str__(self) -> str:
        """String representation of the molecule.

        Examples:
            >>> m = Mol('CCO', name='ethanol')
            >>> print(m)

        Returns:
            str: string representation.
        """
        return f"<Mol({self.smiles} name={self.name} conformers={self.num_confs})>"
    

    def __hash__(self) -> str:
        """Hashed SMILES string of the molecule.

        When you compare two objects using the `==` operator, Python first checks 
        if their hash values are equal. If they are different, the objects are 
        considered unequal, and the __eq__ method is not called.
        The return value of `__hash__` method is also used as dictionary keys or set elements.

        Examples:
            >>> m1 == m2

        Returns:
            str: hashed SMILES string.
        """
        return hash(self.smiles)
    

    def __eq__(self, other: Self) -> bool:
        """True if `other` Mol is identical with this Mol.

        It compares InChIKeys.

        Examples:
            >>> m1 == m2

        Args:
            other (object): other Mol object.

        Returns:
            bool: True if identical.
        """
        return self.InChIKey == other.InChIKey
    

    def __iter__(self) -> Iterator:
        """Yields an iterator of conformers of the molecule.

        Examples:
            >>> for conformer in mol:
            >>>     print(conformer.name)

        Yields:
            Iterator: conformers of the molecule.
        """
        return iter(self.confs)
    

    def __next__(self) -> Conf:
        """Next conformer of the molecule.

        Returns:
            Conf: Conf object of one of conformers of the molecule.
        """
        return next(self.confs)
    

    def __getitem__(self, index: int | slice) -> Conf | Self:
        """Conformer object of conformers of the molecule with given index or slice of indexes.

        Examples:
            >>> first_conformer = mol[0]

        Args:
            index (int | slice): index for conformers.

        Returns:
            Conf or Mol(copy) with conformers specified by index.
        """
        assert self.num_confs > 0, "no conformers"
        
        if isinstance(index, slice):
            new_object = self.copy()
            new_object.confs = new_object.confs[index]
            return new_object
        
        else:
            return self.confs[index]

    ##################################################
    ### Properties
    ##################################################

    @property
    def num_confs(self) -> int:
        """Returns the total number of conformers.

        Returns:
            int: total count of conformers.
        """
        return len(self.confs)
    

    @property
    def charge(self) -> int:
        """Returns molecular formal charge

        Returns:
            int: molecular formal charge
        """
        return rdmolops.GetFormalCharge(self.rdmol)
    

    @property
    def symbols(self) -> list[str]:
        """Returns the element symbols.

        Returns:
            list: list of element symbols.
        """
        return [atom.GetSymbol() for atom in self.rdmol.GetAtoms()]


    @property
    def numbers(self) -> list[int]:
        """Returns the atomic numbers.

        Returns:
            list: list of atomic numbers.
        """
        return [atom.GetAtomicNum() for atom in self.rdmol.GetAtoms()]
    

    @property
    def molblock(self) -> str:
        """Returns MolBlock"""
        return Chem.MolToMolBlock(self.rdmol)
    

    @property
    def is_stereo_specified(self) -> bool:
        """Check if the molecule is stereo-specified at tetrahedral atom and double bond.

        This function uses `Chem.FindPotentialStereo()` function which returns a list of `elements`.
        Explanation of the elements:
            element.type: 
                whether the element is a stereocenter ('stereoAtom') or a stereobond ('stereoBond')
                - Atom_Octahedral
                - Atom_SquarePlanar
                - *Atom_Tetrahedral*
                - Atom_TrigonalBipyramidal
                - Bond_Atropisomer
                - Bond_Cumulene_Even
                - *Bond_Double*m.
                - Unspecified 

            element.centeredOn:
                The atom or bond index where the stereochemistry is centered.
            
            element.specified:
                A boolean indicating whether the stereochemistry at that location 
                is explicitly specified in the molecule.
                values = {
                    0: rdkit.Chem.rdchem.StereoSpecified.Unspecified, 
                    1: rdkit.Chem.rdchem.StereoSpecified.Specified, 
                    2: rdkit.Chem.rdchem.StereoSpecified.Unknown,
                    }
            
            element.descriptor:
                A descriptor that can be used to identify the type of stereochemistry (e.g., 'R', 'S', 'E', 'Z').
                - Bond_Cis = rdkit.Chem.StereoDescriptor.Bond_Cis
                - Bond_Trans = rdkit.Chem.StereoDescriptor.Bond_Trans
                - NoValue = rdkit.Chem.StereoDescriptor.NoValue
                - Tet_CCW = rdkit.Chem.StereoDescriptor.Tet_CCW
                - Tet_CW = rdkit.Chem.StereoDescriptor.Tet_CW

        Returns:
            bool: True if stereo-specified.
        """
        stereos = []
        for element in Chem.FindPotentialStereo(self.rdmol):
            if element.type == Chem.StereoType.Atom_Tetrahedral:
                stereos.append(element.specified == Chem.StereoSpecified.Specified)
            elif element.type == Chem.StereoType.Bond_Double :
                bond = self.rdmol.GetBondWithIdx(element.centeredOn)
                if bond.GetBeginAtom().GetSymbol() == 'N' or bond.GetEndAtom().GetSymbol() == 'N':
                    continue
                else:
                    stereos.append(element.specified == Chem.StereoSpecified.Specified)
        
        # note all([]) returns True
        return all(stereos)


    @property
    def ring_bond_stereo_info(self) -> list[tuple]:
        """Returns double bond and cis/trans stereochemistry information.

        Returns:
            list[tuple]: [(element.centeredOn, element.descriptor), ...]
        """
        stereo_info = Chem.FindPotentialStereo(self.rdmol)
        info = []
        for element in stereo_info:
            if element.type == Chem.StereoType.Bond_Double:
                if self.rdmol.GetBondWithIdx(element.centeredOn).IsInRing():
                    info.append((element.centeredOn, element.descriptor))
        
        return info
    
    @property
    def num_stereoisomers(self) -> int:
        """Counts number of all possible stereoisomers ignoring the current stereochemistry.

        Returns:
            int: number of stereoisomers.
        """

        ring_bond_stereo_info = self.ring_bond_stereo_info
        mol = self.copy()
        # remove stereochemistry
        mol = mol.remove_stereo()
        rdmols = enumerate_stereoisomers(mol.rdmol)
        # ring bond stereo is not properly enumerated
        # cis/trans information is lost if stereochemistry is removed,
        # which cannot be enumerated by EnumerateStereoisomers() function
        # so enumerate_ring_bond_stereoisomers() is introduced
        if len(ring_bond_stereo_info) > 0:
            ring_cis_trans = []
            for rdmol in rdmols:
                ring_cis_trans += enumerate_ring_bond_stereoisomers(rdmol,
                                                        ring_bond_stereo_info,
                                                        override=True)
            if len(ring_cis_trans) > 0:
                rdmols = ring_cis_trans
        
        unique_rdmols = set([Chem.MolToSmiles(rdmol) for rdmol in rdmols])
        
        return len(unique_rdmols)
    
        
    ##################################################
    ### Pipeline Functions (returns Self)
    ##################################################

    def from_molblock(self, molblock: str, compressed: bool = False) -> Self:
        """Initialize a new Mol object from MolBlock.

        Args:
            molblock (str): MolBlock string

        Raises:
            ValueError: invalid MolBlock

        Returns:
            Self: self.
        """
        
        if compressed:
            molblock = decompress_string(molblock)
        
        molecule = Chem.MolFromMolBlock(molblock)
        
        try:
            self.rdmol, _ = clean_2d(molecule, reset_isotope=True, remove_H=True)
            self.smiles = Chem.MolToSmiles(self.rdmol)
            self.confs = [Conf(x) for x in _]
        except:
            raise ValueError(f'Mol() Error: invalid MolBlock string')

        assert self.smiles and self.rdmol, "Mol() Error: invalid molecule"
        
        name = self.rdmol.GetProp('_Name')

        rdDepictor.Compute2DCoords(self.rdmol)

        try:
            self.name = str(name)
        except:
            self.name = 'untitled'
        
        self.rdmol.SetProp('_Name', self.name) # _Name can't be None        
        self.InChIKey = generate_inchi_key(self.rdmol)
        self.props.update({
            'aka' : [], # <-- to be set by MolLibr.unique()
            'atoms' : self.rdmol.GetNumAtoms(), # hydrogens not excluded?
            'charge': rdmolops.GetFormalCharge(self.rdmol),
            "nrb" : Descriptors.NumRotatableBonds(self.rdmol),
            })
        
        return self
    

    def deserialize(self, serialized: str) -> Self:
        """De-serialize the information and build a new Mol object.

        Example:
            serialized = mol1.serialize()
            mol2 = Mol().deserialize(serialized)

        Args:
            serialized (str): serialized string.

        Returns:
            Self: modified self.
        """
        data = rdworks.utils.deserialize(serialized)

        self.name = data['name']
        self.smiles = data['smiles'] # isomeric SMILES, no H
        
        try:
            self.rdmol = Chem.MolFromMolBlock(data['molblock'])
        except KeyError:
            # backward-compatible
            self.rdmol = Chem.MolFromSmiles(data['smiles']) # for 2D depiction
            self.rdmol.SetProp('_Name', self.name)
    
        self.InChIKey = data['InChIKey']
        self.props = data['props']
        self.confs = [Conf().deserialize(_) for _ in data['confs']] # for 3D conformers (iterable)
        
        return self

    
    def copy(self) -> Self:
        """Returns a copy of self.

        Returns:
            a copy of self.
        """
        return copy.deepcopy(self)


    def rename(self, 
               prefix: str = '', 
               sep: str = '/', 
               start: int = 1) -> Self:
        """Updates name and conformer names.
        
        The first conformer name is {prefix}{sep}{start}

        Args:
            prefix (str, optional): prefix of the name. Defaults to ''.
            sep (str, optional): separtor betwween prefix and serial number. Defaults to '/'.
            start (int, optional): first serial number. Defaults to 1.

        Returns:
            Self: modified self.
        """
        if prefix :
            self.name = prefix
            self.rdmol.SetProp('_Name', prefix)
        
        # update conformer names
        num_digits = len(str(self.num_confs)) # ex. '100' -> 3
        for (serial, conf) in enumerate(self.confs, start=start):
            serial_str = str(serial)
            while len(serial_str) < num_digits:
                serial_str = '0' + serial_str
            conf.rename(f'{self.name}{sep}{serial_str}')
        
        return self
    

    def qed(self, 
            properties: list[str] = ['QED', 'MolWt', 'LogP', 'TPSA', 'HBD']) -> Self:
        """Updates quantitative estimate of drug-likeness (QED) and other descriptors.

        Args:
            properties (list[str], optional): Defaults to ['QED', 'MolWt', 'LogP', 'TPSA', 'HBD'].

        Raises:
            KeyError: if property key is unknown.

        Returns:
            Self: modified self.
        """
        props_dict = {}
        for k in properties:
            try:
                props_dict[k] = rd_descriptor_f[k](self.rdmol)
            except:
                raise KeyError(f'qed() Error: unknown property {k}')
        self.props.update(props_dict)
        
        return self
    

    def remove_stereo(self) -> Self:
        """Removes stereochemistry. 

        Examples:
            >>> m = Mol("C/C=C/C=C\\C", "double_bond")
            >>> m.remove_stereo().smiles == "CC=CC=CC"

        Returns:
            Self: modified self.
        """
        # keep the original stereo info. for ring double bond
        Chem.RemoveStereochemistry(self.rdmol)
        Chem.AssignStereochemistry(self.rdmol, 
                                   cleanIt=False, 
                                   force=False, 
                                   flagPossibleStereoCenters=False)
        self.smiles = Chem.MolToSmiles(self.rdmol)

        return self
   

    def make_confs(self, n: int = 50, method: str = 'ETKDG', **kwargs) -> Self:
        """Generates 3D conformers.

        Args:
            n (int, optional): number of conformers to generate. Defaults to 50.
            method (str, optional): conformer generation method.
                Choices are `ETKDG`, `CONFORGE`. Defaults to 'ETKDG'.

        Returns:
            Self: modified self.

        Reference:
            T. Seidel, C. Permann, O. Wieder, S. M. Kohlbacher, T. Langer, 
            High-Quality Conformer Generation with CONFORGE: Algorithm and Performance Assessment. 
            J. Chem. Inf. Model. 63, 5549-5570 (2023).
        """
        verbose = kwargs.get('verbose', False)

        self.confs = []

        if method.upper() == 'ETKDG':
            rdmol_H = Chem.AddHs(self.rdmol, addCoords=True) # returns a copy with hydrogens added
            conf_ids = rdDistGeom.EmbedMultipleConfs(rdmol_H, numConfs=n, params=self.ETKDG_params)
            for rdConformer in rdmol_H.GetConformers():
                # number of atoms should match with conformer(s)
                rdmol_conf = Chem.Mol(rdmol_H)
                rdmol_conf.RemoveAllConformers()
                rdmol_conf.AddConformer(Chem.Conformer(rdConformer))
                conf = Conf(rdmol_conf)
                self.confs.append(conf)
        
        elif method.upper() == 'CONFORGE':
            mol = CDPL.Chem.parseSMILES(self.smiles)
            # create and initialize an instance of the class ConfGen.ConformerGenerator which
            # will perform the actual conformer ensemble generation work
            conf_gen = CDPL.ConfGen.ConformerGenerator()
            conf_gen.settings.timeout = 60 * 1000  # 60 sec.
            conf_gen.settings.minRMSD = 0.5
            conf_gen.settings.energyWindow = 20.0 # kcal/mol(?)
            conf_gen.settings.maxNumOutputConformers = n
            # dictionary mapping status codes to human readable strings
            status_to_str = { 
                CDPL.ConfGen.ReturnCode.UNINITIALIZED                  : 'uninitialized',
                CDPL.ConfGen.ReturnCode.TIMEOUT                        : 'max. processing time exceeded',
                CDPL.ConfGen.ReturnCode.ABORTED                        : 'aborted',
                CDPL.ConfGen.ReturnCode.FORCEFIELD_SETUP_FAILED        : 'force field setup failed',
                CDPL.ConfGen.ReturnCode.FORCEFIELD_MINIMIZATION_FAILED : 'force field structure refinement failed',
                CDPL.ConfGen.ReturnCode.FRAGMENT_LIBRARY_NOT_SET       : 'fragment library not available',
                CDPL.ConfGen.ReturnCode.FRAGMENT_CONF_GEN_FAILED       : 'fragment conformer generation failed',
                CDPL.ConfGen.ReturnCode.FRAGMENT_CONF_GEN_TIMEOUT      : 'fragment conformer generation timeout',
                CDPL.ConfGen.ReturnCode.FRAGMENT_ALREADY_PROCESSED     : 'fragment already processed',
                CDPL.ConfGen.ReturnCode.TORSION_DRIVING_FAILED         : 'torsion driving failed',
                CDPL.ConfGen.ReturnCode.CONF_GEN_FAILED                : 'conformer generation failed',
                }
            
            # We have to create a temporary file and re-read it for storing individual conformers.
            tmp_dir = os.environ.get('TMPDIR', '/tmp')
            tmp_name = next(tempfile._get_candidate_names()) + '.sdf'
            tmp_filename = os.path.join(tmp_dir, tmp_name)
            
            writer = CDPL.Chem.MolecularGraphWriter(tmp_filename, "sdf" )
            # SB - io.StringIO does not work with Chem.MolecularGraphWriter()            
             
            try:
                # prepare the molecule for conformer generation
                CDPL.ConfGen.prepareForConformerGeneration(mol) 
                # generate the conformer ensemble
                status = conf_gen.generate(mol)             
                # if successful, store the generated conformer ensemble as
                # per atom 3D coordinates arrays (= the way conformers are represented in CDPKit)
                if status == CDPL.ConfGen.ReturnCode.SUCCESS or status == CDPL.ConfGen.ReturnCode.TOO_MUCH_SYMMETRY:
                    # TOO_MUCH_SYMMETRY: output ensemble may contain duplicates
                    conf_gen.setConformers(mol)
                    writer.write(mol)  
                    writer.close()                          
                else:
                    raise RuntimeError('Error: conformer generation failed: %s' % status_to_str[status])
            except Exception as e:
                raise RuntimeError('Error: conformer generation failed: %s' % str(e))
            
            # tmpfile is automatically closed here but kept, as delete=False was set
            
            with Chem.SDMolSupplier(tmp_filename, sanitize=True, removeHs=False) as sdf:
                self.confs = [ Conf(m) for m in sdf if m is not None ]
            
            # tmpfile is not deleted here because delete=False
            # we should remove the file when it is no longer needed
            os.remove(tmp_filename)

        # energy evaluations for ranking
        calculator = kwargs.get('calculator', 'MMFF94')
        self.singlepoint_confs(calculator, **kwargs)
        
        # set relative energy, E_rel(kcal/mol)
        sort_by = 'E_tot(kcal/mol)'
        self.confs = sorted(self.confs, key=lambda c: c.props[sort_by]) # ascending order
        lowest_energy = self.confs[0].props[sort_by]
        for conf in self.confs:
            conf.props.update({"E_rel(kcal/mol)": conf.props[sort_by] - lowest_energy})

        # rename conformers
        self = self.rename()

        if verbose:
            rot_bonds = rd_descriptor_f['RotBonds'](self.rdmol)
            nrb_suggested = int(8.481 * (rot_bonds **1.642))
            logger.info(f"make_confs() rotatable bonds {rot_bonds} (suggested conformers {nrb_suggested}) generated {self.num_confs}")
            logger.info(f"make_confs() updated potential energies E_tot(kcal/mol) and E_rel(kcal/mol) by {calculator}")

        return self


    def singlepoint_confs(self, 
                          calculator: str | Callable, 
                          batchsize_atoms: int = 0,
                          water: str | None = None) -> Self:
        """Evaluates potential energy of each conformer without geometry optimization.

        It sets `E_tot(kcal/mol)` property for each conformer.
        
        Args:
            calculator (str | Callable): MMFF94 (= MMFF), MMFF94s, UFF, xTB or ASE calculator.
                `MMFF94` or `MMFF` - Intended for general use, including organic molecules and proteins, 
                    and primarily relies on data from quantum mechanical calculations. 
                    It's often used in molecular dynamics simulations.
                `MMFF94s` - A "static" variant of MMFF94, with adjusted parameters for out-of-plane 
                    bending and dihedral torsions to favor planar geometries for specific nitrogen atoms. 
                    This makes it better suited for geometry optimization studies where a static, 
                    time-averaged structure is desired. The "s" stands for "static".
                `UFF` - UFF refers to the "Universal Force Field," a force field model used for 
                    molecular mechanics calculations. It's a tool for geometry optimization, 
                    energy minimization, and exploring molecular conformations in 3D space. 
                    UFF is often used to refine conformers generated by other methods, 
                    such as random conformer generation, to produce more physically plausible 
                    and stable structures.
            fmax (float, optional): fmax for the calculator convergence. Defaults to 0.05.
            max_iter (int, optional): max iterations for the calculator. Defaults to 1000.
            batchsize_atoms (int, optional): maximum number of atoms in a single batch.
                Setting any number smaller than conf.natoms to disable batch optimization.
                Defaults to 0.

        Args for xTB calculator:
            water (str, optional): water solvation model (choose 'gbsa' or 'alpb')
                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
                gbsa: generalized Born (GB) model with Surface Area contributions.
                Defaults to None.

        Returns:
            Self: modified self.
        """
        if isinstance(calculator, Callable) and batchsize_atoms >= self.confs[0].natoms:
            self.confs = batch_singlepoint(self.confs, calculator, batchsize_atoms)
        else:
            self.confs = [conf.singlepoint(calculator, water=water) for conf in self.confs]
        
        return self
    

    def optimize_confs(self, 
                       calculator: str | Callable = 'MMFF94', 
                       fmax: float = 0.05, 
                       max_iter: int = 1000,
                       batchsize_atoms: int = 0,
                       water: str | None = None
                       ) -> Self:
        """Optimizes 3D geometry of conformers.

        Args:
            calculator (str | Callable): MMFF94 (= MMFF), MMFF94s, UFF, or ASE calculator.
                `MMFF94` or `MMFF` - Intended for general use, including organic molecules and proteins, 
                    and primarily relies on data from quantum mechanical calculations. 
                    It's often used in molecular dynamics simulations.
                `MMFF94s` - A "static" variant of MMFF94, with adjusted parameters for out-of-plane 
                    bending and dihedral torsions to favor planar geometries for specific nitrogen atoms. 
                    This makes it better suited for geometry optimization studies where a static, 
                    time-averaged structure is desired. The "s" stands for "static".
                `UFF` - UFF refers to the "Universal Force Field," a force field model used for 
                    molecular mechanics calculations. It's a tool for geometry optimization, 
                    energy minimization, and exploring molecular conformations in 3D space. 
                    UFF is often used to refine conformers generated by other methods, 
                    such as random conformer generation, to produce more physically plausible 
                    and stable structures.
            fmax (float, optional): fmax for the calculator convergence. Defaults to 0.05.
            max_iter (int, optional): max iterations for the calculator. Defaults to 1000.
            batchsize_atoms (int, optional): max number of atoms in one batch. 
                Defaults to 16384(=16*1024). Disable batch optimization if zero or negative.
        
        Args for xTB calculator:
            water (str, optional): water solvation model (choose 'gbsa' or 'alpb')
                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
                gbsa: generalized Born (GB) model with Surface Area contributions.
                Defaults to None.
        
        Returns:
            Self: modified self.
        """
        if isinstance(calculator, Callable) and batchsize_atoms >= self.confs[0].natoms:
            self.confs = batch_optimize(self.confs, calculator, batchsize_atoms)
        else:
            self.confs = [ conf.optimize(calculator, fmax, max_iter, water=water) for conf in self.confs ]

        return self


    def sort_confs(self, calculator: str | Callable | None = None, **kwargs) -> Self:
        """Sorts by `E_tot(kcal/mol)` or `E_tot(eV)` and sets `E_rel(kcal/mol)`. 

        Args:
            calculator (str | Callable | None): MMFF94 (= MMFF), MMFF94s, UFF, or ASE calculator.
                `MMFF94` or `MMFF` - Intended for general use, including organic molecules and proteins, 
                    and primarily relies on data from quantum mechanical calculations. 
                    It's often used in molecular dynamics simulations.
                `MMFF94s` - A "static" variant of MMFF94, with adjusted parameters for out-of-plane 
                    bending and dihedral torsions to favor planar geometries for specific nitrogen atoms. 
                    This makes it better suited for geometry optimization studies where a static, 
                    time-averaged structure is desired. The "s" stands for "static".
                `UFF` - UFF refers to the "Universal Force Field," a force field model used for 
                    molecular mechanics calculations. It's a tool for geometry optimization, 
                    energy minimization, and exploring molecular conformations in 3D space. 
                    UFF is often used to refine conformers generated by other methods, 
                    such as random conformer generation, to produce more physically plausible 
                    and stable structures.
        Raises:
            KeyError: if `E_tot(eV)` or `E_tot(kcal/mol)` is not defined.

        Returns:
            Self: modified self.
        """
        verbose = kwargs.get('verbose', False)

        if calculator is not None:
            # re-calculate potential energies
            if verbose :
                logger.info(f"sort_cons() calculate potential energy by {calculator}")
            self.singlepoint_confs(calculator, **kwargs)

        if all(['E_tot(kcal/mol)' in conf.props for conf in self.confs]):
            sort_by = 'E_tot(kcal/mol)'
            conversion = 1.0

        elif all(['E_tot(eV)' in conf.props for conf in self.confs]):
            sort_by = 'E_tot(eV)'
            conversion = ev2kcalpermol  # eV to kcal/mol
        
        else:
            raise KeyError(f'sort_confs() requires `E_tot(eV)` or `E_tot(kcal/mol)` property')
        
        # ascending order
        self.confs = sorted(self.confs, key=lambda c: c.props[sort_by])

        if self.num_confs > 0:
            E_min = self.confs[0].props[sort_by]
            for conf in self.confs:
                E_rel = conversion * (conf.props[sort_by] - E_min)
                conf.props.update({"E_rel(kcal/mol)": E_rel})
        
        return self
    

    def align_confs(self, method: str = 'rigid_fragment') -> Self:
        """Aligns all conformers to the first conformer.

        Args:
            method (str, optional): alignment method: 
                `rigid_fragment`, `CrippenO3A`, `MMFFO3A`, `best_rms`.
                Defaults to `rigid_fragment`.

        Returns:
            Self: modified self.
        """

        if self.num_confs < 2: 
            # nothing to do
            return self

        if method == 'rigid_fragment': 
            indices = rigid_fragment_indices(self.confs[0].rdmol)[0] # 3D and H, largest fragment
            atomMap = [(i, i) for i in indices]
            for i in range(1, self.num_confs):
                rmsd = rdMolAlign.AlignMol(prbMol=self.confs[i].rdmol, refMol=self.confs[0].rdmol, atomMap=atomMap)
                # rdMolAlign.AlignMol does not take symmetry into account
                # but we will use atom indices for alignment anyway.
                # If atomMap is not given, AlignMol() will attempt to generate atomMap by
                # substructure matching.
            self.is_confs_aligned = True
        
        elif method == 'CrippenO3A':
            crippen_ref_contrib = rdMolDescriptors._CalcCrippenContribs(self.confs[0].rdmol)
            for i in range(1, self.num_confs):
                crippen_prb_contrib = rdMolDescriptors._CalcCrippenContribs(self.confs[i].rdmol)
                crippen_O3A = rdMolAlign.GetCrippenO3A(prbMol=self.confs[i].rdmol,
                                                       refMol=self.confs[0].rdmol, 
                                                       prbCrippenContribs=crippen_prb_contrib, 
                                                       refCrippenContribs=crippen_ref_contrib, 
                                                       )
                crippen_O3A.Align()
                # crippen_O3A.Score()
            self.is_confs_aligned = True

        elif method == 'MMFFO3A':
            mmff_ref_params = AllChem.MMFFGetMoleculeProperties(self.confs[0].rdmol)
            for i in range(1, self.num_confs):
                mmff_prb_params = AllChem.MMFFGetMoleculeProperties(self.confs[i].rdmol)
                mmff_O3A = rdMolAlign.GetO3A(prbMol=self.confs[i].rdmol, 
                                             refMol=self.confs[0].rdmol, 
                                             prbPyMMFFMolProperties=mmff_prb_params, 
                                             refPyMMFFMolProperties=mmff_ref_params, 
                                            )
                mmff_O3A.Align()
                # mmff_O3A.Score()
            self.is_confs_aligned = True

        elif method == 'best_rms':
            for i in range(1, self.num_confs):
                # symmetry-aware alignment / speed can be improved by removing Hs
                rmsd = rdMolAlign.GetBestRMS(prbMol=self.confs[i].rdmol, refMol=self.confs[0].rdmol)
            self.is_confs_aligned = True

        return self
    

    def cluster_confs(self, 
                      method: str = 'QT', 
                      threshold: float = 1.0, 
                      sort: str = 'size',
                      symmetry_aware : bool = True) -> Self:
        """Clusters all conformers and sets cluster properties.
        
        Conformers are expected to be aligned.

        Following cluster properties will be added: `cluster`, `cluster_mean_energy`, 
            `cluster_median_energy`, `cluster_IQR_energy`, `cluster_size`, `cluster_centroid` (True or False)                  
        
        `RCKMeans` algorithm is unreliable and not supported for now.
        
        Args:
            method (str, optional): clustering algorithm:
                `Butina`, 
                `QT`,
                `NMRCLUST`,
                `DQT`,
                `BitQT`, 
                `DynamicTreeCut`, 
                `AutoGraph`.
                Defaults to `QT`.
            threshold (float, optional): RMSD threshold of a cluster. Defaults to 1.0.
            sort (str, optional): sort cluster(s) by mean `energy` or cluster `size`. 
                Defaults to `size`.
            symmetry_aware (bool, optional): whether to use symmetry-aware rmsd.
                Defaults to True.

        Raises:
            NotImplementedError: if unsupported method is requested.

        Returns:
            Self: modified self.
        """
        if not self.is_confs_aligned:
            self = self.align_confs()

        if method != 'DQT': # rmsd of x,y,z coordinates (non-H)
            conf_rdmols_noH = [Chem.RemoveHs(Chem.Mol(conf.rdmol)) for conf in self.confs]
            # copies are made for rmsd calculations to prevent coordinates changes
            lower_triangle_values = []
            if symmetry_aware:
                # symmetry-aware RMSD calculation using spyrmsd
                for i in range(1, self.num_confs):
                    mol_i = spyrmsd_Molecule.from_rdkit(self.confs[i].rdmol)
                    mol_js = [spyrmsd_Molecule.from_rdkit(self.confs[j].rdmol) for j in range(i)]
                    rmsd_list = spyrmsd_wrapper(mol_i, mol_js, symmetry=True, minimize=False)
                    # it just calculates rmsd without aligning (minimize=False)
                    lower_triangle_values += rmsd_list
            else:
                for i in range(1, self.num_confs): # number of conformers
                    for j in range(i):
                        # rdMolAlign.GetBestRMS takes symmetry into account
                        # removed hydrogens to speed up
                        best_rms = rdMolAlign.GetBestRMS(prbMol=conf_rdmols_noH[i], refMol=conf_rdmols_noH[j])
                        lower_triangle_values.append(best_rms)
        
        else: # rmsd (radian) of dihedral angles
            torsion_angle_atom_indices_list = self.get_torsion_angle_atoms()
            # [(5, 4, 3, 1), ...]
            # symmmetry-related equivalence is not considered
            torsions = []
            for conf in self.confs:
                t_radians = []
                for (i, j, k, l) in torsion_angle_atom_indices_list:
                    t_radians.append(
                        rdMolTransforms.GetDihedralRad(conf.rdmol.GetConformer(), i, j, k, l))
                torsions.append(np.array(t_radians))
            # torsions: num.confs x num.torsions
            N = len(torsions)
            lower_triangle_values = []
            for i in range(N):
                for j in range(i):
                    rad_diff = np.fmod(torsions[i] - torsions[j], 2.0*np.pi)
                    rmsd = np.sqrt(np.sum(rad_diff**2)/N)
                    # np.max(np.absolute(rad_diff))
                    lower_triangle_values.append(rmsd)

        cluster_assignment = None
        centroid_indices = None

        if method == 'Butina':
            clusters = Butina.ClusterData(data=lower_triangle_values, 
                                          nPts=self.num_confs,
                                          distThresh=threshold,
                                          isDistData=True,
                                          reordering=True)
            cluster_assignment = [None,] * self.num_confs
            centroid_indices = []
            for cluster_idx, indices in enumerate(clusters):
                for conf_idx in indices:
                    cluster_assignment[conf_idx] = cluster_idx
                centroid_indices.append(indices[0])

        elif method == 'QT': 
            # my implementation of the original QT algorithm
            # tighter than Butina
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = QT(symm_matrix, threshold)
        
        elif method == 'NMRCLUST': 
            # looser than Butina
            # does not require threshold
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = NMRCLUST(symm_matrix)

        elif method == 'DQT':
            # issues with symmetry related multiplicities
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = QT(symm_matrix, threshold)

        elif method == 'BitQT':
            # supposed to produce identical result as QT but it does not
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = BitQT(symm_matrix, threshold)
        
        elif method == 'DynamicTreeCut':
            # often collapses into single cluster. so not very useful.
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = DynamicTreeCut(symm_matrix)
        
        # elif method == 'RCKmeans':
        #     # buggy
        #     symm_matrix = convert_tril_to_symm(lower_triangle_values)
        #     cluster_assignment, centroid_indices = RCKmeans(symm_matrix)
        
        elif method == 'AutoGraph':
            # not reliable
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = AutoGraph(symm_matrix)

        else:
            raise NotImplementedError(f'{method} clustering is not implemented yet.')

        # cluster_assignment: ex. [0,1,0,0,2,..]
        # centroid_indices: ex. [10,5,..] i.e. centroids of clusters 0 and 1 are 10 and 5, respectively.

        if cluster_assignment is not None and centroid_indices is not None:
            cluster_raw_data = defaultdict(list)
            for conf_idx, cluster_idx in enumerate(cluster_assignment):
                cluster_raw_data[cluster_idx].append(conf_idx)
            cluster_list = []
            for i, k in enumerate(sorted(cluster_raw_data.keys())):
                energies = [self.confs[conf_idx].props['E_rel(kcal/mol)'] for conf_idx in cluster_raw_data[k]]
                mean_energy = np.mean(energies)
                median_energy = np.median(energies)
                q75, q25 = np.percentile(energies, [75, 25])
                iqr_energy = q75 - q25 # interquartile range (IQR)
                cluster_list.append({'confs' : cluster_raw_data[k],
                                     'centroid' : centroid_indices[i], # conformer index
                                     'size' : len(cluster_raw_data[k]),
                                     'mean_energy' : mean_energy,
                                     'median_energy' : median_energy,
                                     'iqr_energy' : iqr_energy,
                                     })
            # sort cluster index
            if sort == 'size':
                cluster_list = sorted(cluster_list, key=lambda x: x['size'], reverse=True)
            
            elif sort == 'energy':
                cluster_list = sorted(cluster_list, key=lambda x: x['median_energy'], reverse=False)
            
            else:
                raise NotImplementedError(f'{sort} is not implemented yet.')

            for cluster_idx, cluster_dict in enumerate(cluster_list, start=1):
                for conf_idx in cluster_dict['confs']:
                    if conf_idx == cluster_dict['centroid']:
                        self.confs[conf_idx].props.update({
                                'cluster' : cluster_idx,
                                'cluster_mean_energy' : cluster_dict['mean_energy'],
                                'cluster_median_energy' : cluster_dict['median_energy'],
                                'cluster_IQR_energy' : cluster_dict['iqr_energy'],
                                'cluster_size' : cluster_dict['size'],
                                'cluster_centroid' : True,
                                })
                    else:
                        self.confs[conf_idx].props.update({
                                'cluster' : cluster_idx,
                                'cluster_mean_energy' : cluster_dict['mean_energy'],
                                'cluster_median_energy' : cluster_dict['median_energy'],
                                'cluster_IQR_energy' : cluster_dict['iqr_energy'],
                                'cluster_size' : cluster_dict['size'],
                                'cluster_centroid' : False,
                                })
        return self


    def drop_confs(self,
                   stereo_flipped: bool = True,
                   unconverged: bool = True,
                   similar: bool | None = None,
                   similar_rmsd: float = 0.3,
                   cluster: bool | None =None,
                   k: int | None = None,
                   window: float | None = None,
                   **kwargs) -> Self:
        """Drop conformers that meet some condition(s).

        Args:
            stereo_flipped (bool): drop conformers whose R/S and cis/trans stereo is unintentionally flipped.
                For example, a trans double bond in a macrocyle can end up with both trans
                and cis isomers in the final optimized conformers.
            unconverged (bool): drop unconverged conformers. see `Converged` property.
            similar (bool, optional): drop similar conformers. see `similar_rmsd`.
            similar_rmsd (float): RMSD (A) below `similar_rmsd` is regarded similar (default: 0.3)
            cluster (bool, optional): drop all except for the lowest energy conformer in each cluster.
            k (int, optional): drop all except for `k` lowest energy conformers.
            window (float, optional): drop all except for conformers within `window` of relative energy.
        
        Examples:
            To drop similar conformers within rmsd of 0.5 A
            >>> mol.drop_confs(similar=True, similar_rmsd=0.5)
            
            To drop conformers beyond 5 kcal/mol
            >>> mol.drop_confs(window=5.0)

        Returns:
            Self: modified self.    
        """

        verbose = kwargs.get('verbose', False)

        reasons = [f'stereo flipped',  
                   f'unconverged', 
                   f'similar({similar_rmsd})',
                   f'cluster(non-centroid)',
                   f'k and/or energy window',
                   ]
        
        w = max([len(s) for s in reasons])
               
        if stereo_flipped and self.num_confs > 0:
            mask = [Chem.MolToSmiles(Chem.RemoveHs(_.rdmol)) == self.smiles for _ in self.confs]
            self.confs = list(itertools.compress(self.confs, mask))
            if verbose:
                logger.info(f'drop_confs() {mask.count(False):3d} {reasons[0]:<{w}} -> {self.num_confs}')
        
        if unconverged and self.num_confs > 0:
            mask = [_.props['Converged'] if 'Converged' in _.props else True for _ in self.confs]
            self.confs = list(itertools.compress(self.confs, mask))
            if verbose:
                logger.info(f'drop_confs() {mask.count(False):3d} {reasons[1]:<{w}} -> {self.num_confs}')

        if similar and self.num_confs > 1:
            # it is observed that there are essentially identical conformers 
            # such as 180-degree ring rotation and there is not minor conformational variations
            # in the RDKit ETKDG generated conformers.
            
            if not self.is_confs_aligned:
                self = self.align_confs()

            # symmetry-aware RMSD calculation using spyrmsd
            lower_triangle_values = []
            for i in range(1, self.num_confs):
                mol_i = spyrmsd_Molecule.from_rdkit(self.confs[i].rdmol)
                mol_js = [spyrmsd_Molecule.from_rdkit(self.confs[j].rdmol) for j in range(i)]
                rmsd_list = spyrmsd_wrapper(mol_i, mol_js, symmetry=True, minimize=False)
                # it just calculates rmsd without aligning (minimize=False)
                lower_triangle_values += rmsd_list
            symm_matrix = convert_tril_to_symm(lower_triangle_values)
            cluster_assignment, centroid_indices = QT(symm_matrix, similar_rmsd)
            mask = [conf_idx in centroid_indices for conf_idx, conf in enumerate(self.confs)]
            # keep the centroid of clusters and drop others
            self.confs = list(itertools.compress(self.confs, mask))
            if verbose:
                logger.info(f'drop_confs() {mask.count(False):3d} {reasons[2]:<{w}} -> {self.num_confs}')

            # note: it will retain the conformers with lower index
            # so, it should be sorted before dropping
            # obj = obj.sort_confs()
            # mask = []
            # retained_confs = []
            # for conf_i in obj.confs:
            #     is_dissimilar = True
            #     for conf_j_rdmol_noH in retained_confs:
            #         # symmetry-aware alignment / removing Hs speeds up the calculation
            #         rmsd = rdMolAlign.GetBestRMS(Chem.RemoveHs(conf_i.rdmol), conf_j_rdmol_noH)
            #         if rmsd < similar_rmsd:
            #             is_dissimilar = False
            #             break
            #     mask.append(is_dissimilar)
            #     if is_dissimilar:
            #         retained_confs.append(Chem.RemoveHs(conf_i.rdmol)) # store a copy of H-removed rdmol
            # obj.confs = list(itertools.compress(obj.confs, mask))
        
        if cluster and self.num_confs > 1:
            # drop non-centroid cluster member(s)
            mask = [_.props['centroid'] if 'centroid' in _.props else True for _ in self.confs]
            self.confs = list(itertools.compress(self.confs, mask))
            if verbose:
                logger.info(f'drop_confs() {mask.count(False):3d} {reasons[3]:<{w}} -> {self.num_confs}')

        if (k or window) and self.num_confs > 0:
            # confs must be sorted by energies
            if not all(['E_rel(kcal/mol)' in _.props for _ in self.confs]):
                self = self.sort_confs(**kwargs)
            if k:
                mask_k = [i < k for i,_ in enumerate(self.confs)]
            else:
                mask_k = [True,] * self.num_confs
            if window:
                mask_window = [_.props['E_rel(kcal/mol)'] < window if 'E_rel(kcal/mol)' in _.props else True for _ in self.confs]
            else:
                mask_window = [True,] * self.num_confs
            # retain conformer(s) that satisfy both k and window conditions
            mask = [(x and y) for (x,y) in zip(mask_k, mask_window)]
            self.confs = list(itertools.compress(self.confs, mask))
            if verbose:
                logger.info(f'drop_confs() {mask.count(False):3d} {reasons[4]:<{w}} -> {self.num_confs}')
        
        return self
            
    
    def compute(self, **kwargs) -> Self:
        """Change settings for parallel computing.

        Args:
            max_workers (int): max number of workers.
            chunksize (int): chunksize of splitted workload.
            progress (bool): whether to show progress bar.

        Returns:
            Self: modified self.
        """
        self.max_workers = kwargs.get('max_workers', self.max_workers)
        self.chunksize = kwargs.get('chunksize', self.chunksize)
        self.progress = kwargs.get('progress', self.progress)
        
        return self
    

    def calculate_torsion_energies(self,
                         calculator: str | Callable = 'MMFF94',
                         torsion_angle_idx: int | None = None,
                         simplify: bool = True, 
                         fmax: float = 0.05,
                         interval: float = 20.0,
                         use_converged_only: bool = True,
                         batchsize_atoms: int = 0,
                         water: str | None = None,
                         ) -> Self:
        """Calculates potential energy profiles for each torsion angle using ASE optimizer.

        It uses the first conformer as a reference.

        Args:
            calculator (str | Callable): 'MMFF', 'UFF', 'xTB' or ASE calculator.
            torsion_angle_idx (int | None): torsion index to calculate. Defaults to None (all).
            simplify (bool, optional): whether to use fragment surrogate. Defaults to True.
            fmax (float, optional): fmax of ASE optimizer. Defaults to 0.05.
            interval (float, optional): interval of torsion angles in degree. Defaults to 15.0.
            use_converged_only (bool, optional): whether to use only converged data. Defaults to True.
            batchsize_atoms (int, optional): maximum number of atoms in a single batch.
                Setting any number smaller than conf.natoms to disable batch optimization.
                Defaults to 0.
        
        Args for xTB calculator:
            water (str, optional): water solvation model (choose 'gbsa' or 'alpb')
                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
                gbsa: generalized Born (GB) model with Surface Area contributions.
                Defaults to None.

        Returns:
            Self: modified self.
        """
        assert self.num_confs > 0, "calculate_torsion_energies() requires at least one conformer"
        
        ref_conf = self.confs[0].copy()
        
        if batchsize_atoms >= ref_conf.natoms:
            assert not isinstance(calculator, str), "Batch optimizer is required for batch calculation"

        ref_conf = ref_conf.calculate_torsion_energies(calculator, 
                                            torsion_angle_idx, 
                                            simplify, 
                                            fmax, 
                                            interval, 
                                            use_converged_only,
                                            water=water,
                                            batchsize_atoms=batchsize_atoms)
        
        self.props['torsion'] = ref_conf.props['torsion']

        return self
    

    def draw(self, 
             coordgen: bool = False,
             rotate: bool = False,
             axis: str = 'z', 
             degree: float = 0.0,
             ) -> Self:
        """Draw molecule in 2D.

        Args:
            coordgen (bool, optional): whether to use `coordgen`. Defaults to False.
            rotate (bool, optional): whether to rotate drawing. Defaults to False.
            axis (str, optional): axis for rotation. Defaults to 'z'.
            degree (float, optional): degree for rotation. Defaults to 0.0.

        Returns:
            Self.
        """
        rdDepictor.SetPreferCoordGen(coordgen)
        rdDepictor.Compute2DCoords(self.rdmol)
        
        if rotate:
            rad = (np.pi/180.0) * degree
            c = np.cos(rad)
            s = np.sin(rad)
            if axis.lower() == 'x':
                rotmat = np.array([
                    [1., 0., 0., 0.],
                    [0., c, -s,  0.],
                    [0., s,  c,  0.],
                    [0., 0., 0., 1.],
                    ])
            elif axis.lower() == 'y':
                rotmat = np.array([
                    [ c,  0., s,  0.],
                    [ 0., 1., 0., 0.],
                    [-s,  0., c,  0.],
                    [ 0., 0., 0., 1.],
                    ])
            elif axis.lower() == 'z':
                rotmat = np.array([
                    [c, -s,  0., 0.],
                    [s,  c,  0., 0.],
                    [0., 0., 1., 0.],
                    [0., 0., 0., 1.],
                    ])
            rdMolTransforms.TransformConformer(
                self.rdmol.GetConformer(), 
                rotmat)        
        
        return self
    

    ##################################################
    ### Endpoint Functions (returns some values)
    ##################################################

    def is_nnp_ready(self,  model: str = 'aimnet2') -> bool:
        """Check if a particular neural network model is applicable to current molecule.

        Args:
            model (str): neural network models: `ANI-2x`, `ANI-2xt`, `AIMNET`

        Raises:
            ValueError: if model is not supported.

        Returns:
            bool: True if applicable.
        """
        if model.lower() in ['ani-2x', 'ani-2xt', 'ani2x', 'ani2xt']:
            if self.props['charge'] != 0:
                return False
            # H, C, N, O, F, S, Cl
            atomic_numbers = [1, 6, 7, 8, 9, 16, 17 ]
        
        elif model.lower() in ['aimnet', 'aimnet2']:
            # H, B, C, N, O, F, Si, P, S, Cl, As, Se, Br, I
            atomic_numbers = [1, 5, 6, 7, 8, 9, 14, 15, 16, 17, 33, 34, 35, 53 ]
        
        else:
            raise ValueError('nnp_ready() supports ANI-2x, ANI-2xt, AIMNet, or AIMNet2')
        
        if all([ a.GetAtomicNum() in atomic_numbers for a in self.rdmol.GetAtoms() ]):
            return True
        else:
            return False
        

    def get_torsion_angle_atoms(self, strict: bool = True) -> list[tuple]:
        """Determine torsion/dihedral angle atoms (i-j-k-l) and rotating group for each rotatable bond (j-k).

        Args:
            strict (bool): whether to exclude amide/imide/ester/acid bonds.

        Returns:
            [(i, j, k, l), ...]
        """
        return [d[:4] for d in get_torsion_angle_atom_indices(self.rdmol, strict)]
    

    def get_similarity(self, other: Self) -> float:
        """Returns Tanimoto similarity with other Mol object.

        Args:
            other (Mol): other Mol object.

        Raises:
            TypeError: if `other` is not Mol object type.

        Returns:
            float: Tanimoto similarity.
        """
        assert isinstance(other, Mol), "similarity() Error: invalid Mol object"
        
        if not self.fp:
            self.fp = self.MFP2.GetFingerprint(self.rdmol)
        
        if not other.fp:
            other.fp = other.MFP2.GetFingerprint(other.rdmol)
        
        return DataStructs.TanimotoSimilarity(self.fp, other.fp)
    

    def is_similar(self, other: Self, threshold: float) -> bool:
        """Check if other molecule is similar within Tanimoto similarity threshold.

        Args:
            other (Mol): other Mol object to compare with.
            threshold (float): Tanimoto similarity threshold.

        Returns:
            bool: True if similar.
        """
        return self.get_similarity(other) >= threshold


    def has_substr(self, substr: str) -> bool:
        """Determine if the molecule has the substructure match.

        Args:
            pattern (str): SMARTS or SMILES.

        Returns:
            bool: True if matches.
        """
        query = Chem.MolFromSmarts(substr)
        return self.rdmol.HasSubstructMatch(query)


    def is_matching(self, terms: str | Path, invert: bool = False) -> bool:
        """Determines if the molecule matches the predefined substructure and/or descriptor ranges.

        invert | terms(~ or !) | effect
        ------ | ------------- | -------------
        True   |     ~         | No inversion
        True   |               | Inversion
        False  |     ~         | Inversion
        False  |               | No inversion

        Args:
            terms (str | Path): 
                substructure SMARTS expression or a path to predefined descriptor ranges.
            invert (bool, optional): whether to invert the result. Defaults to False.

        Returns:
            bool: True if matches.
        """
        if isinstance(terms, Path):
            path = terms.as_posix()

        elif isinstance(terms, str):
            if terms.startswith('~') or terms.startswith('!'):
                terms = terms.replace('~','').replace('!','')
                invert = (invert ^ True)
            try:
                path = Path(terms) # test if terms points to a xml file
                assert path.is_file()
            except:
                path = get_predefined_xml(terms)
        else:
            print(list_predefined_xml())
            return False
        
        (lterms, combine) = parse_xml(path)
        mask = []
        for (name, smarts, lb, ub) in lterms:
            if smarts:
                query= Chem.MolFromSmarts(smarts)
                if len(self.rdmol.GetSubstructMatches(query)) > 0:
                    mask.append(True)
                else:
                    mask.append(False)
            else: # descriptor lower and upper bounds
                if name not in self.props:
                    val = rd_descriptor_f[name](self.rdmol)
                    self.props.update({name: val})
                else:
                    val = self.props[name]
                # return if lower and upper boundaries are satisfied
                if ((not lb) or (val >= lb)) and ((not ub) or (val <= ub)):
                    mask.append(True)
                else:
                    mask.append(False)
            if combine.lower() == 'or' and any(mask):
                # early termination if any term is satisfied
                return invert ^ True # XOR(^) inverts only if invert is True
        
        if combine.lower() == 'and' and all(mask):
            return invert ^ True
        
        return invert ^ False
    

    def to_sdf(self, confs: bool = False, props: bool = True) -> str:
        """Returns strings of SDF output.

        Args:
            confs (bool, optional): whether to include conformers. Defaults to False.
            props (bool, optional): whether to include properties. Defaults to True.

        Returns:
            str: strings of SDF output.
        """
        buf = StringIO()
        with Chem.SDWriter(buf) as f:
            if confs:
                for conf in self.confs:
                    rdmol = Chem.Mol(conf.rdmol)
                    rdmol.SetProp('_Name', conf.name)
                    if props:
                        # molcule props.
                        for k,v in self.props.items():
                            rdmol.SetProp(k, str(v))
                        # conformer props.
                        for k,v in conf.props.items():
                            rdmol.SetProp(k, str(v))
                    f.write(rdmol)
            else:
                rdmol = Chem.Mol(self.rdmol)
                rdmol.SetProp('_Name', self.name)
                if props:
                    for k,v in self.props.items():
                        rdmol.SetProp(k, str(v))
                f.write(rdmol)
        
        return buf.getvalue()
       

    def to_png(self,
               width: int = 300, 
               height: int = 300,
               legend: str = '', 
               atom_index: bool = False, 
               highlight_atoms: list[int] | None = None,
               highlight_bonds: list[int] | None = None,
               redraw: bool = False,
               coordgen: bool = False,
               trim: bool = True,
               ) -> Image.Image:
        """Draw 2D molecule in PNG format.

        Args:
            width (int, optional): width. Defaults to 300.
            height (int, optional): height. Defaults to 300.
            legend (str, optional): legend. Defaults to ''.
            atom_index (bool, optional): whether to show atom index. Defaults to False.
            highlight_atoms (list[int] | None, optional): atom(s) to highlight. Defaults to None.
            highlight_bonds (list[int] | None, optional): bond(s) to highlight. Defaults to None.
            redraw (bool, optional): whether to redraw. Defaults to False.
            coordgen (bool, optional): whether to use coordgen. Defaults to False.
            trim (bool, optional): whether to trim white margins. Default to True.
        
        Returns:
            Image.Image: output PIL Image object.
        """
        
        return render_png(self.rdmol,
                          width = width, 
                          height = height, 
                          legend = legend, 
                          atom_index = atom_index, 
                          highlight_atoms = highlight_atoms,
                          highlight_bonds = highlight_bonds,
                          redraw = redraw, 
                          coordgen = coordgen, 
                          trim = trim)

    def to_svg(self, 
               width: int = 300, 
               height: int = 300,
               legend: str = '', 
               atom_index: bool = False, 
               highlight_atoms: list[int] | None = None,
               highlight_bonds: list[int] | None = None,
               redraw: bool = False,
               coordgen: bool = False,
               optimize: bool = True,
               ) -> str:
        """Draw 2D molecule in SVG format.

        Examples:
            For Jupyternotebook, wrap the output with SVG:

            >>> from IPython.display import SVG
            >>> SVG(libr[0].to_svg())

        Args:
            width (int, optional): width. Defaults to 300.
            height (int, optional): height. Defaults to 300.
            legend (str, optional): legend. Defaults to ''.
            atom_index (bool, optional): whether to show atom index. Defaults to False.
            highlight_atoms (list[int] | None, optional): atom(s) to highlight. Defaults to None.
            highlight_bonds (list[int] | None, optional): bond(s) to highlight. Defaults to None.
            redraw (bool, optional): whether to redraw. Defaults to False.
            coordgen (bool, optional): whether to use coordgen. Defaults to False.
            optimize (bool, optional): whether to optimize SVG string. Defaults to True.

        Returns:
            str: SVG string
        """
        return render_svg(self.rdmol, 
                          width = width, 
                          height = height, 
                          legend = legend, 
                          atom_index = atom_index, 
                          highlight_atoms = highlight_atoms,
                          highlight_bonds = highlight_bonds,
                          redraw = redraw, 
                          coordgen = coordgen, 
                          optimize = optimize)
    

    def to_plot_data_torsion_angle_vs_energy(self, idx: int | None = None) -> dict:
        """Returns plot data for torsion angle vs energy.

        For seaborn plot,

            ```py
            data = self.props['torsion'][torsion_angle_idx]
            df = pd.DataFrame({ax: data[ax] for ax in ['angle', 'E_rel(kcal/mol)']})
            
            plt.figure(**kwargs)
            plt.clf()  # Clear the current figure to prevent overlapping plots

            sns.set_theme()
            sns.color_palette("tab10")
            sns.set_style("whitegrid")

            if len(df['angle']) == len(df['angle'].drop_duplicates()):
                g = sns.lineplot(x="angle",  
                                y="E_rel(kcal/mol)", 
                                data=df, 
                                marker='o', 
                                markersize=10)
            else:
                g = sns.lineplot(x="angle",
                                y="E_rel(kcal/mol)", 
                                data=df, 
                                errorbar=('ci', 95),
                                err_style='bars',
                                marker='o',
                                markersize=10)
            g.xaxis.set_major_locator(ticker.MultipleLocator(30))
            g.xaxis.set_major_formatter(ticker.ScalarFormatter())
            if df["E_rel(kcal/mol)"].max() > upper_limit:
                g.set(title=self.name,
                    xlabel='Dihedral Angle (degree)', 
                    ylabel='Relative Energy (Kcal/mol)',
                    xlim=(-190, 190),
                    ylim=(-1.5, upper_limit))
            elif df["E_rel(kcal/mol)"].max() < zoomin_limit:
                g.set(title=self.name,
                    xlabel='Dihedral Angle (degree)', 
                    ylabel='Relative Energy (Kcal/mol)',
                    xlim=(-190, 190),
                    ylim=(-1.5, zoomin_limit))
            else:
                g.set(title=self.name, 
                    xlabel='Dihedral Angle (degree)', 
                    ylabel='Relative Energy (Kcal/mol)',
                    xlim=(-190, 190),)
            g.tick_params(axis='x', rotation=30)
            
            if svg:
                buf = StringIO()
                plt.savefig(buf, format='svg', bbox_inches='tight')
                plt.close() # prevents duplicate plot outputs in Jupyter Notebook
                svg_string = buf.getvalue()
                # optimize SVG string
                scour_options = {
                    'strip_comments': True,
                    'strip_ids': True,
                    'shorten_ids': True,
                    'compact_paths': True,
                    'indent_type': 'none',
                }
                svg_string = scourString(svg_string, options=scour_options)

                return svg_string
            
            else:
                buf = BytesIO()
                plt.savefig(buf, format='png', bbox_inches='tight')
                plt.close() # prevents duplicate plot outputs in Jupyter Notebook
                buf.seek(0)
                img = Image.open(buf)
                plt.imshow(img)
                plt.axis('off') # Optional: remove axes
                plt.show()
            ```

            Args:
                idx (int, optional) - 0-based torsion angle index. 
                    Defaults to None (all torsion angles).
            
            Returns:
                {'idx': list, 'angle': list, 'E_rel(kcal/mol)': list}

        """
        if isinstance(idx, int):
            _dict = self.props['torsion'][idx]
            angles, energies = _dict['angle'], _dict['E_rel(kcal/mol)']
            assert len(angles) == len(energies)
            data = {'idx':[idx,]*len(angles), 'angle':angles, 'E_rel(kcal/mol)':energies}
        else:
            data = {'idx':[], 'angle':[], 'E_rel(kcal/mol)':[]}
            for idx, _dict in enumerate(self.props['torsion']): # list
                angles, energies = _dict['angle'], _dict['E_rel(kcal/mol)']
                assert len(angles) == len(energies)
                data['idx'].extend([idx,]*len(angles))
                data['angle'].extend(angles)
                data['E_rel(kcal/mol)'].extend(energies)
        
        return data


    # def plot_torsion_energies(self, 
    #                           torsion_angle_idx: int,
    #                           svg: bool = False,
    #                           upper_limit: float = 35.0,
    #                           zoomin_limit: float = 5.0,
    #                           **kwargs,
    #                           ) -> str | None:
    #     """Plot torsion energies.

    #     Args:
    #         torsion_angle_idx (int): torsion data to plot.
    #         svg (bool, optional): whether to return SVG strings. Defaults to False.
    #         upper_limit (float, optional): upper limit for E_rel(kcal/mol). Defaults to 35.0.
    #         zoomin_limit (float, optional): lower limit for E_rel(kcal/mol). Defaults to 5.0.
    #         **kwargs: matplotlib.pyplot.plt.figure options.

    #     Returns:
    #         SVG strings or None for Jupyter Notebook.
    #     """
    #     data = self.props['torsion'][torsion_angle_idx]
    #     df = pd.DataFrame({ax: data[ax] for ax in ['angle', 'E_rel(kcal/mol)']})
        
    #     plt.figure(**kwargs)
    #     plt.clf()  # Clear the current figure to prevent overlapping plots

    #     sns.set_theme()
    #     sns.color_palette("tab10")
    #     sns.set_style("whitegrid")

    #     if len(df['angle']) == len(df['angle'].drop_duplicates()):
    #         g = sns.lineplot(x="angle",  
    #                          y="E_rel(kcal/mol)", 
    #                          data=df, 
    #                          marker='o', 
    #                          markersize=10)
    #     else:
    #         g = sns.lineplot(x="angle",
    #                          y="E_rel(kcal/mol)", 
    #                          data=df, 
    #                          errorbar=('ci', 95),
    #                          err_style='bars',
    #                          marker='o',
    #                          markersize=10)
    #     g.xaxis.set_major_locator(ticker.MultipleLocator(30))
    #     g.xaxis.set_major_formatter(ticker.ScalarFormatter())
    #     if df["E_rel(kcal/mol)"].max() > upper_limit:
    #         g.set(title=self.name,
    #               xlabel='Dihedral Angle (degree)', 
    #               ylabel='Relative Energy (Kcal/mol)',
    #               xlim=(-190, 190),
    #               ylim=(-1.5, upper_limit))
    #     elif df["E_rel(kcal/mol)"].max() < zoomin_limit:
    #         g.set(title=self.name,
    #               xlabel='Dihedral Angle (degree)', 
    #               ylabel='Relative Energy (Kcal/mol)',
    #               xlim=(-190, 190),
    #               ylim=(-1.5, zoomin_limit))
    #     else:
    #         g.set(title=self.name, 
    #               xlabel='Dihedral Angle (degree)', 
    #               ylabel='Relative Energy (Kcal/mol)',
    #               xlim=(-190, 190),)
    #     g.tick_params(axis='x', rotation=30)
        
    #     if svg:
    #         buf = StringIO()
    #         plt.savefig(buf, format='svg', bbox_inches='tight')
    #         plt.close() # prevents duplicate plot outputs in Jupyter Notebook
    #         svg_string = buf.getvalue()
    #         # optimize SVG string
    #         scour_options = {
    #             'strip_comments': True,
    #             'strip_ids': True,
    #             'shorten_ids': True,
    #             'compact_paths': True,
    #             'indent_type': 'none',
    #         }
    #         svg_string = scourString(svg_string, options=scour_options)

    #         return svg_string
        
    #     else:
    #         buf = BytesIO()
    #         plt.savefig(buf, format='png', bbox_inches='tight')
    #         plt.close() # prevents duplicate plot outputs in Jupyter Notebook
    #         buf.seek(0)
    #         img = Image.open(buf)
    #         plt.imshow(img)
    #         plt.axis('off') # Optional: remove axes
    #         plt.show()
        

    # def to_html(self, htmlbody: bool = False, contents: str = 'torsion') -> str:
    #     """Returns HTML text of dihedral energy profile.

    #     Args:
    #         htmlbody (bool, optional): whether to wrap around with `<html><body>`. Defaults to False.

    #     Returns:
    #         str: HTML text.
    #     """
    #     HTML = ''
    #     if htmlbody:
    #         HTML = '<html><body>'

    #     if contents.lower() == 'torsion':
    #         # start of content
    #         HTML += f'<h1 style="text-align:left">{self.name}</h1>'
    #         HTML += '<table>'
    #         for torsion_angle_idx, dictdata in enumerate(self.props['torsion']):
    #             ijkl = dictdata['indices']
    #             ijkl_str = '-'.join([str(i) for i in ijkl])
    #             svg_mol = self.to_svg(highlight_atoms=ijkl, atom_index=True)
    #             svg_plot = self.plot_torsion_energies(torsion_angle_idx=torsion_angle_idx, svg=True)
    #             frag = dictdata.get('frag', None)
    #             if frag is not None:
    #                 frag = Chem.MolFromMolBlock(frag)
    #                 pqrs = dictdata['frag_indices']
    #                 pqrs_str = '-'.join([str(i) for i in pqrs])
    #                 svg_frag = render_svg(frag, highlight_atoms=pqrs, atom_index=True)
    #                 HTML += f'<tr><td>{ijkl_str}</td><td>{svg_mol}</td>'
    #                 HTML += f'<td>{pqrs_str}<td>{svg_frag}</td><td>{svg_plot}</td></tr>'
    #             else:
    #                 HTML += f'<tr><td>{ijkl_str}</td><td>{svg_mol}</td><td>{svg_plot}</td></tr>'
    #         HTML += '</table>'
    #         HTML += '<hr style="height:2px;border-width:0;color:gray;background-color:gray">'
    #         # end of content

    #     if htmlbody:
    #         HTML += '</body></html>'

    #     return HTML


    def dumps(self, key: str = "", decimals: int = 2) -> str:
        """Returns JSON dumps of properties.

        Args:
            key (str | None): key for a subset of properties. Defaults to None.
            decimals (int, optional): decimal places for float numbers. Defaults to 2.

        Returns:
            str: JSON dumps.
        """
        props = recursive_round(self.props, decimals)

        if key:
            return json.dumps({key:props[key]})
        
        return json.dumps(props)


    def serialize(self, decimals: int = 3) -> str:
        """Serialize information necessary to rebuild a Mol object.

        Args:
            decimals (int, optional): number of decimal places for float data type. Defaults to 2.

        Returns:
            str: serialized string for json.loads()
        """
        return rdworks.utils.serialize({
            'name'  : self.name,
            'smiles': self.smiles,
            'molblock': self.molblock,
            'InChIKey': self.InChIKey,
            'props' : recursive_round(self.props, decimals),
            'confs' : [conf.serialize(decimals=decimals) for conf in self.confs],
            })


    ##################################################
    ### Report Functions (returns None)
    ##################################################

    def report_stereo(self) -> None:
        """Report stereochemistry information for debug"""
        num_chiral_centers = rdMolDescriptors.CalcNumAtomStereoCenters(self.rdmol)
        # Returns the total number of atomic stereocenters (specified and unspecified)
        num_unspecified_chiral_centers = rdMolDescriptors.CalcNumUnspecifiedAtomStereoCenters(self.rdmol)
        print(f"chiral centers = unspecified {num_unspecified_chiral_centers} / total {num_chiral_centers}")
        print(f"stereogenic double bonds =")
        for element in Chem.FindPotentialStereo(self.rdmol):
            # element.type= Atom_Octahedral, Atom_SquarePlanar, Atom_Tetrahedral, 
            #               Atom_TrigonalBipyramidal, 
            #               Bond_Atropisomer, Bond_Cumulene_Even, Bond_Double, 
            #               Unspecified 
            if element.type == Chem.StereoType.Bond_Double:
                bond = self.rdmol.GetBondWithIdx(element.centeredOn)
                atom1 = bond.GetBeginAtom().GetSymbol()
                atom2 = bond.GetEndAtom().GetSymbol()
                is_nitrogen = (atom1 == 'N' or atom2 == 'N')
                print(f'  {element.type} bond: {element.centeredOn}', end=' ')
                print(f'ring: {bond.IsInRing()} N: {is_nitrogen}', end=' ')
            elif element.type == Chem.StereoType.Atom_Tetrahedral:
                print(f'  {element.type} atom: {element.centeredOn}', end=' ')
                print(f'atoms {list(element.controllingAtoms)}', end=' ')
            print(f'{element.specified} {element.descriptor}') # type: Chem.StereoDescriptor


    def report_props(self) -> None:
        """Report properties"""
        if self.props:
            print(f"Properties({len(self.props)}):")
            fixed_width = max([len(k) for k in self.props]) + 4
            for k,v in self.props.items():
                while len(k) <= fixed_width:
                    k = k + ' '
                print(f"  {k} {v}")
        else:
            print(f"Properties: None")


from rdkit import Chem
from rdkit.Chem import AllChem, DataStructs


def distance_function(i: int, j: int, fps: list) -> float:
    """Calculate Tanimoto distance between two fingerprints."""
    return 1.0 - DataStructs.TanimotoSimilarity(fps[i], fps[j])


def calculate_fingerprints(rdmols: list[Chem.Mol], fp_type: str='morgan', radius:int=2, n_bits:int=2048) -> list:
    """
    Calculate molecular fingerprints for a list of SMILES.
    
    Args:
        smiles_list: List of SMILES strings
        fp_type: Type of fingerprint ('morgan', 'rdkit', 'maccs')
        radius: Radius for Morgan fingerprints
        n_bits: Number of bits for fingerprints
    
    Returns:
        List of fingerprint objects
    """
    
    fps = []
    
    for mol in rdmols:
        if mol is None:
            fps.append(None)
            continue
            
        if fp_type == 'morgan':
            fp = AllChem.GetMorganFingerprintAsBitVect(mol, radius, nBits=n_bits)
        elif fp_type == 'rdkit':
            fp = Chem.RDKFingerprint(mol, fpSize=n_bits)
        elif fp_type == 'maccs':
            fp = AllChem.GetMACCSKeysFingerprint(mol)
        else:
            raise ValueError(f"Unknown fingerprint type: {fp_type}")
        
        fps.append(fp)
    
    return fps





import numpy as np

from rdkit import Chem
from rdworks.xtb.wrapper import GFN2xTB


def get_torsion_angle_atom_indices(rdmol:Chem.Mol, strict:bool=True) -> list[tuple]:
    """Determine dihedral angle atoms (a-b-c-d) and rotating group for each rotatable bond (b-c).

    Args:
        rdmol (Chem.Mol): molecule
        strict (bool): whether to exclude amide/imide/ester/acid bonds.

    Returns:
        [   (a, b, c, d, rot_atom_indices, fix_atom_indices), 
            (a, b, c, d, rot_atom_indices, fix_atom_indices), 
            ...,
        ]
    """
    # https://github.com/rdkit/rdkit/blob/1bf6ef3d65f5c7b06b56862b3fb9116a3839b229/rdkit/Chem/Lipinski.py#L47%3E
    # https://github.com/rdkit/rdkit/blob/de602c88809ea6ceba1e8ed50fd543b6e406e9c4/Code/GraphMol/Descriptors/Lipinski.cpp#L108
    if strict :
        # excludes amide/imide/ester/acid bonds
        rotatable_bond_pattern = Chem.MolFromSmarts(
            (
            "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])("
            "[CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])&!$([#7,O,S!D1]-!@[CD3]="
            "[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])&!$([#7!D1]-!@[CD3]=[N+])]-,:;!@[!$"
            "(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])(["
            "CH3])[CH3])]"
            )
        )
    else:
        rotatable_bond_pattern = Chem.MolFromSmarts('[!$(*#*)&!D1]-&!@[!$(*#*)&!D1]')
    
    rotatable_bonds = rdmol.GetSubstructMatches(rotatable_bond_pattern)
    
    torsion_angle_atom_indices = []

    # small rings (n=3 or 4)
    small_rings = [ r for r in list(rdmol.GetRingInfo().AtomRings()) if len(r) < 5 ]
    # ex. = [(1, 37, 35, 34, 3, 2), (29, 28, 30)]

    forbidden_terminal_nuclei = [1, 9, 17, 35, 53] # H,F,Cl,Br,I

    for (b_idx, c_idx) in rotatable_bonds:
        # determine a atom ``a`` that define a dihedral angle 
        a_candidates = []
        for neighbor in rdmol.GetAtomWithIdx(b_idx).GetNeighbors():
            neighbor_idx = neighbor.GetIdx()
            if neighbor_idx == c_idx:
                continue
            neighbor_atomic_num = neighbor.GetAtomicNum()
            if neighbor_atomic_num not in forbidden_terminal_nuclei:
                a_candidates.append((neighbor_atomic_num, neighbor_idx))
        
        if not a_candidates:
            continue
        
        (a_atomic_num, a_idx) = sorted(a_candidates, key=lambda x: (x[0], -x[1]), reverse=True)[0]

        # is a-b in a small ring (n=3 or 4)?
        is_in_small_ring = False
        for small_ring in small_rings:
            if (a_idx in small_ring) and (b_idx in small_ring):
                is_in_small_ring = True
                break
        
        if is_in_small_ring:
            continue

        # determine a atom ``d`` that define a dihedral angle
        d_candidates = []
        for neighbor in rdmol.GetAtomWithIdx(c_idx).GetNeighbors():
            neighbor_idx = neighbor.GetIdx()
            if (neighbor_idx == b_idx):
                continue
            neighbor_atomic_num = neighbor.GetAtomicNum()
            if neighbor_atomic_num not in forbidden_terminal_nuclei:
                d_candidates.append((neighbor_atomic_num, neighbor_idx))
        
        if not d_candidates:
            continue
        
        (d_atomic_num, d_idx) = sorted(d_candidates, key=lambda x: (x[0], -x[1]), reverse=True)[0]

        # is c-d in a small ring?
        is_in_small_ring = False
        for small_ring in small_rings:
            if (c_idx in small_ring) and (d_idx in small_ring):
                is_in_small_ring = True
                break
        
        if is_in_small_ring:
            continue

        # determine a group of atoms to be rotated
        # https://ctr.fandom.com/wiki/Break_rotatable_bonds_and_report_the_fragments
        em = Chem.EditableMol(rdmol)
        em.RemoveBond(b_idx, c_idx)
        fragmented = em.GetMol()
        (frag1, frag2) = Chem.GetMolFrags(fragmented, asMols=False) # returns tuple of tuple
        hac1 = sum([ 1 for i in frag1 if rdmol.GetAtomWithIdx(i).GetAtomicNum() > 1 ])
        hac2 = sum([ 1 for i in frag2 if rdmol.GetAtomWithIdx(i).GetAtomicNum() > 1 ])
        
        # smaller fragment will be rotated and must contain at least three heavy atoms
        if min(hac1, hac2) >= 3:
            (frag_rot, frag_fix) = sorted([(hac1, frag1), (hac2, frag2)])
            torsion_angle_atom_indices.append((a_idx, b_idx, c_idx, d_idx, frag_rot[1], frag_fix[1]))

    return torsion_angle_atom_indices




def find_atoms_at_bond_distance(rdmol: Chem.Mol,
                                start_atom_idx: int,
                                distance: int) -> list[int]:
    """Finds atoms at a specific bond distance from a starting atom.

    Args:
        mol: An RDKit Mol object.
        start_atom_idx: The index of the starting atom.
        distance: The desired bond distance.

    Returns:
        A list of atom indices at the specified distance
    """
    assert start_atom_idx < rdmol.GetNumAtoms(), "start_atom_idx out of range."

    found_atoms = []
    visited = set()

    def dfs(curr_atom_idx: int, curr_bond_dist: int):
        if curr_bond_dist == distance:
            found_atoms.append(curr_atom_idx)
            return

        visited.add(curr_atom_idx)
        curr_atom = rdmol.GetAtomWithIdx(curr_atom_idx)
        for next_atom in curr_atom.GetNeighbors():
            next_atom_idx = next_atom.GetIdx()
            if next_atom_idx not in visited:
                dfs(next_atom_idx, curr_bond_dist + 1)

        # Backtrack
        visited.remove(curr_atom_idx)

    dfs(start_atom_idx, 0)

    return found_atoms


def get_bond_distance(rdmol: Chem.Mol, start_atom_idx: int) -> dict:
    """Get bonds distance from a given atom.

    Args:
        mol: An RDKit Mol object.
        start_atom_idx: The index of the starting atom.
        distance: The desired bond distance.

    Returns:
        A list of atom indices at the specified distance
    """
    assert start_atom_idx < rdmol.GetNumAtoms(), "start_atom_idx out of range."

    bond_distance = {}
    visited = set()

    def dfs(curr_atom_idx: int, curr_bond_dist: int):
        if curr_bond_dist in bond_distance:
            bond_distance[curr_bond_dist].append(curr_atom_idx)
        else:
            bond_distance[curr_bond_dist] = [curr_atom_idx]

        visited.add(curr_atom_idx)
        curr_atom = rdmol.GetAtomWithIdx(curr_atom_idx)
        for next_atom in curr_atom.GetNeighbors():
            next_atom_idx = next_atom.GetIdx()
            if next_atom_idx not in visited:
                dfs(next_atom_idx, curr_bond_dist + 1)

        # Backtrack
        visited.remove(curr_atom_idx)

    dfs(start_atom_idx, 0)

    return bond_distance


def find_bonds_to_prune(rdmol: Chem.Mol, 
                        torsion_indices: tuple,
                        bond_dist_threshold: int = 4,
                        bond_order_threshold: float = 1.75,
                        electronegative: list[int] = [7, 8, 9, 17, 35],
                        ) -> dict[int, list[int]]:
    """Find pruning candidate bonds from a given atom to construct fragment.

    Rules for a candidate bond to break:
        
        For (i-j-k-l) torsion,

        1. NOT (bond distance from j or k < 4)
        2. NOT (bond order > 1.75)
        3. NOT (Pauling electronegativity of any of bond atoms > 2.9)

    Args:
        mol: An RDKit Mol object.
        start_atom_idx: The index of the starting atom.
        distance: The desired bond distance.

    Pauling electronegativity:
        ```py
        from mendeleev import element
        for i in range(1, 119):  # 118 is the highest atomic number known
            el = element(i)
            if isinstance(el.en_pauling, float) and el.en_pauling > 2.9:
                print(f"Element {i}: {el.symbol} {el.atomic_number} {el.en_pauling}")
        ```
        Element 7: N 7 3.04
        Element 8: O 8 3.44
        Element 9: F 9 3.98
        Element 17: Cl 17 3.16
        Element 35: Br 35 2.96

    Returns:
        A list of atom indices at the specified distance
    """

    (i, j, k, l) = torsion_indices[:4]
    
    dist_from_j = get_bond_distance(rdmol, j)
    dist_from_k = get_bond_distance(rdmol, k)
    
    # sum(,[]) flattens a list of list
    forbidden  = sum([v for d, v in dist_from_j.items() if d < bond_dist_threshold], [])
    forbidden += sum([v for d, v in dist_from_k.items() if d < bond_dist_threshold], [])
    forbidden = set(forbidden)

    start_atom_idx = k # either j or k yields the same result

    found_bonds = {}
    visited = set()

    def ordered(p: int, q: int) -> list[int]:
        """Returns a list of atom indices by bond distance.

        Args:
            p (int): atom index
            q (int): atom index

        Returns:
            list[int]: (atom index closer to the torsion angle, the other)
        """
        dist_p = []
        dist_q = []
        for d, indices in dist_from_j.items():
            if p in indices:
                dist_p.append(d)
            if q in indices:
                dist_q.append(d)
        for d, indices in dist_from_k.items():
            if p in indices:
                dist_p.append(d)
            if q in indices:
                dist_q.append(d)
        if sum(dist_p) < sum(dist_q):
            return [p, q]
        else:
            return [q, p]

    def dfs(curr_atom_idx: int, bond_dist: int):
        """Depth-first recursive search of bonded atoms.

        Args:
            curr_atom_idx (int): atom index.
            bond_dist (int): bond distance.
        """
        curr_atom = rdmol.GetAtomWithIdx(curr_atom_idx)
        visited.add(curr_atom_idx)
        for next_atom in curr_atom.GetNeighbors():
            next_atom_idx = next_atom.GetIdx()
            bond = rdmol.GetBondBetweenAtoms(curr_atom_idx, next_atom_idx)
            n1 = curr_atom.GetAtomicNum()
            n2 = next_atom.GetAtomicNum()
            # forbidden (rule 1)
            too_close = (curr_atom_idx in forbidden) and (next_atom_idx in forbidden)
            # bond order (rule 2)
            bond_order = not (bond.GetBondTypeAsDouble() > bond_order_threshold)
            # Pauling electronegativity (rule 3)
            bond_pauling = not ((n1 in electronegative) or (n2 in electronegative))
            if (bond_dist >= bond_dist_threshold) and (not too_close) \
                and (not bond.IsInRing()) and bond_order and bond_pauling:
                # determine which atom has shorter bond distance to the torsion angle (j or k)                
                found_bonds[bond.GetIdx()] = ordered(curr_atom_idx, next_atom_idx)
                return
            if next_atom_idx not in visited:
                dfs(next_atom_idx, bond_dist + 1)
        # Backtrack
        visited.remove(curr_atom_idx)

    dfs(start_atom_idx, 0)

    return found_bonds


def get_fragment_idx(parent: Chem.Mol,  
                     indices: tuple, 
                     fragment: Chem.Mol) -> tuple:
    """Get fragment atom indices corresponding to given parent indices.

    It uses 3D coordinates to find matching atoms between parent and fragment.
    In comparison with the MCS-based method `get_fragment_idx_with_mcs()`,
            0 elapsed=0.0006455129478126764 sec.
            1 elapsed=0.0005964740412309766
            2 elapsed=0.0005442029796540737
            3 elapsed=0.000652436981908977
            4 elapsed=0.0006737819639965892
            5 elapsed=0.0004481689538806677
            6 elapsed=0.00035582599230110645
            7 elapsed=0.0003812289796769619
            8 elapsed=0.000359484925866127
            9 elapsed=0.0002818549983203411
            10 elapsed=0.000247497926466167
            11 elapsed=0.0003651580773293972

    Args:
        parent (Chem.Mol): rdkit Chem.Mol object.
        parent_indices (tuple): parent atom indices to map within the MCS.
        fragment (Chem.Mol): fragment originated from the parent.

    Returns:
        dict[int, int]: { parent_atom_index : fragment_atom_index, ...}
    """
    parent_xyz = parent.GetConformer().GetPositions() # numpy.ndarray
    frag_xyz = fragment.GetConformer().GetPositions() # numpy.ndarray
    qpos = [parent_xyz[i] for i in indices]
    
    return tuple(j for q in qpos for j, f in enumerate(frag_xyz) if np.array_equal(f, q))



def get_fragment_idx_with_mcs(parent: Chem.Mol, 
                              indices: tuple, 
                              fragment: Chem.Mol) -> tuple:
    """Get fragment atom indices corresponding to given parent indices.

    Warning:
        It uses MCS and can be extremely slow sometimes.
        For example, below are the elapsed times for 12 torsion angles of atorvastatin:
            0 elapsed=5.525973221054301 sec. **
            1 elapsed=1.9143556850031018 *
            2 elapsed=3.145250838017091 *
            3 elapsed=9.390580283012241 **
            4 elapsed=89.97735002799891 ***
            5 elapsed=0.19022215204313397
            6 elapsed=0.013428106089122593
            7 elapsed=0.023345661000348628
            8 elapsed=0.023358764010481536
            9 elapsed=0.0007965450640767813
            10 elapsed=0.0008196790004149079
            11 elapsed=0.04075543500948697

    Args:
        parent (Chem.Mol): rdkit Chem.Mol object.
        parent_indices (tuple): parent atom indices to map within the MCS.
        fragment (Chem.Mol): fragment originated from the parent.

    Returns:
        dict[int, int]: { parent_atom_index : fragment_atom_index, ...}
    """
    mcs_result = Chem.rdFMCS.FindMCS([parent, fragment])
    mcs_mol = Chem.MolFromSmarts(mcs_result.smartsString)
    
    parent_matches = parent.GetSubstructMatches(mcs_mol)
    frag_matches = fragment.GetSubstructMatches(mcs_mol)

    indices_idx = None
    for parent_matched_indices in parent_matches:
        # It is possible to have more than one matches (i.e. methyl rotation).
        # However, even if there are more than one matches, the parent indices
        # should be the same.
        if indices_idx is None:
            indices_idx = {x: parent_matched_indices.index(x) for x in indices}
        else:
            assert all([indices_idx[x] == parent_matched_indices.index(x) for x in indices])

    indices_map = None
    for frag_matched_indices in frag_matches:
        # it is possible to have more than one matches (i.e. methyl rotation)
        if indices_map is None:
            indices_map = {x : frag_matched_indices[indices_idx[x]] for x in indices}
        else:
            assert all([indices_map[x] == frag_matched_indices[indices_idx[x]] for x in indices])
                
    return tuple([indices_map[x] for x in indices])



def create_fragment_on_bonds(rdmol: Chem.Mol, bonds: dict, cap: bool = True) -> Chem.Mol | None:
    """Create a fragment that preserves defined atoms.

    Args:
        rdmol (Chem.Mol): input molecule.
        bonds (dict): {bond_index : (preserved_atom_index, removed_atom_index), ...}
        cap (bool): whether to cap the dummy atom(s) with hydrogen(s)

    Returns:
        Chem.Mol: resulting fragment molecule.
    """
    fragments = Chem.FragmentOnBonds(rdmol, list(bonds))
    preserved_atoms = { preserved for bond_idx, (preserved, removed) in bonds.items() }
    for fragment_indices, fragment_mol in zip(Chem.GetMolFrags(fragments), 
                                              Chem.GetMolFrags(fragments, asMols=True)):
        if preserved_atoms.issubset(set(fragment_indices)):
            if cap:
                # cap dummy atoms with hydrogens
                for atom in fragment_mol.GetAtoms():
                    if atom.GetAtomicNum() == 0: 
                        atom.SetAtomicNum(1)

            return fragment_mol
        
    return None


def create_torsion_fragment(rdmol: Chem.Mol, 
                            torsion_indices: tuple,
                            wbo_tolerance: float = 0.03) -> tuple[Chem.Mol, list[int], bool, bool]:
    """Create a close surrogate fragment that captures the PES of the intended torsion.
    
    Fragmentation aims to preserve the local chemical environment around the targeted torsion
    while increase calculation speed and potential complications. To avoid oversimplification
    and inaccurate approximation, two strategies are combined:
        - fragment candidates are generated by a set of reasonably empirical rules        
        - further filtered by Wiberg bond order (WBO) calculated by semi-empirical QM. It has 
        been shown that the Wiberg bond order (WBO) provides a fast and robust measure of 
        whether a torsion profile has been disrupted by fragmentation. Any fragment that causes
        WBO difference larger than 0.03 will be excluded.

    Args:
        rdmol (Chem.Mol): molecule.
        torsion_indices (tuple): (i, j, k, l, atoms to be rotated, atoms to be fixed)

    Returns:
        (Chem.Mol: fragment molecule, 
        list[int]: fragment indices, 
        bool: True if fragmented, 
        bool: True if WBO filtering is used)

    References:
        https://pubs.acs.org/doi/10.1021/acs.jcim.2c01153
        https://www.biorxiv.org/content/10.1101/2020.08.27.270934v2
    """
    (i, j, k, l) = torsion_indices[:4]

    candidates = find_bonds_to_prune(rdmol, torsion_indices)
    
    if not candidates:
        # no fragmentation
        return (rdmol, torsion_indices, False, False)

    # fragmented
    WBO_filtered = False
    if GFN2xTB.is_ready():
        # filter candidate(s) by Wiberg bond order (WBO) if xTB is available
        jk = tuple(sorted([j, k]))
        wbo_passed_candidates = {}
        parent = GFN2xTB(rdmol).singlepoint()
        assert hasattr(parent, 'wbo'), "create_torsion_fragment() Error: no wbo for parent"
        for bond_idx, (p, q) in candidates.items():
            frag_single_break = create_fragment_on_bonds(rdmol, {bond_idx: (p, q)})
            fragment = GFN2xTB(frag_single_break).singlepoint()
            assert hasattr(fragment, 'wbo'), "create_torsion_fragment() Error: no wbo for fragment"
            # WBO difference at the torsion angle bond
            frag_jk = get_fragment_idx(rdmol, jk, frag_single_break)
            frag_jk = tuple(sorted(frag_jk))
            if abs(fragment.wbo[frag_jk] - parent.wbo[jk]) < wbo_tolerance:
                wbo_passed_candidates[bond_idx] = (p, q)
        frag_multi_breaks = create_fragment_on_bonds(rdmol, wbo_passed_candidates)
        WBO_filtered = True
    else:
        # skip WBO filtering
        frag_multi_breaks = create_fragment_on_bonds(rdmol, candidates)

    frag_indices = get_fragment_idx(rdmol, (i, j, k, l), frag_multi_breaks)
    
    return (frag_multi_breaks, frag_indices, True, WBO_filtered)


from rdworks import Mol
from rdworks.utils import recursive_round


def test_align_and_cluster_confs():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs()
    mol = mol.drop_confs(similar=True, similar_rmsd=0.5, window=15.0)
    mol = mol.sort_confs().align_confs().cluster_confs().rename()
    mol.to_sdf(confs=True) # string output

def test_make_confs():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs(method='ETKDG')
    assert mol.num_confs > 1
    mol = mol.make_confs(method='CONFORGE')
    assert mol.num_confs > 1


def test_singlepoint():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs(n=10, method='ETKDG')
    mol = mol.singlepoint_confs(calculator='MMFF94')
    assert all([_.props.get('E_tot(kcal/mol)') is not None for _ in mol.confs])


def test_optimize_confs():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs().optimize_confs(calculator='MMFF94')


def test_workflow():
    state_mol = Mol('Cc1nc2cc(Cl)nc(Cl)c2nc1C', 'A-1250')
    state_mol = state_mol.make_confs(n=50, method='ETKDG')
    state_mol = state_mol.drop_confs(similar=True, similar_rmsd=0.3)
    state_mol = state_mol.sort_confs().rename()
    state_mol = state_mol.align_confs(method='rigid_fragment')
    state_mol = state_mol.cluster_confs('QT', threshold=1.0, sort='energy', symmetry_aware=True)
    print(state_mol.name, {k:v for k,v in state_mol.props.items()})
    for conf in state_mol.confs:
        conf.props = recursive_round(conf.props, decimals=2)
        print(conf.name, {k:v for k,v in conf.props.items()})


def test_protonate_deprotonate():
    mol = Mol('c1ccc(CNc2ncnc3ccccc23)cc1') # SAMPL6 SM07
    mol = mol.make_confs(n=1)
    # SMILES: c1ccc(CNc2ncnc3ccccc23)cc1
    # Formal charge: 0
    # Origin: None
    # Transformation: None
    # Ionizable sites:
    # - atom_idx=  5, atom=  N, q= +0, hs= 1, pr= 1, de= 1, acid_base= B:A:A, name= Amine:Amide:Amide vinylogue
    # - atom_idx=  7, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Aza-aromatics
    # - atom_idx=  9, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Aza-aromatics

    conf = mol.confs[0].copy()
    assert conf.positions.shape == (31, 3)
    assert conf.charge == 0
    
    conf = conf.protonate([7])
    assert conf.positions.shape == (32, 3)
    assert conf.charge == 1
    
    conf = mol.confs[0].copy()
    conf = conf.protonate([7,9])
    assert conf.positions.shape == (33, 3)
    assert conf.charge == 2
    
    conf = mol.confs[0].copy()
    conf = conf.deprotonate([5])
    assert conf.positions.shape == (30, 3)
    assert conf.charge == -1


def test_from_molblock():
    mol = Mol('c1ccc(CNc2ncnc3ccccc23)cc1') # SAMPL6 SM07
    mol = mol.make_confs(n=1)
    conf = mol.confs[0].copy()
    mol2 = Mol().from_molblock(conf.molblock)


def test_serialization():
    smiles = 'CC(C)C1=C(C(=C(N1CC[C@H](C[C@H](CC(=O)O)O)O)C2=CC=C(C=C2)F)C3=CC=CC=C3)C(=O)NC4=CC=CC=C4'
    name = 'atorvastatin'
    mol = Mol(smiles, name)
    assert mol.num_confs == 0
    assert mol.name == name
    mol = mol.make_confs(n=10)
    assert mol.num_confs == 10
    serialized = mol.serialize()
    rebuilt = Mol().deserialize(serialized)
    assert rebuilt.num_confs == 10
    assert rebuilt.name == name
    assert rebuilt == mol
    print(mol.confs[0].positions.shape)